"""Defines :class:`CorrelationSurface` and functions to find and build them from a ZX graph."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from functools import cached_property, reduce
from itertools import chain
from operator import xor
from typing import TYPE_CHECKING, NamedTuple

import stim

from tqec.utils.enums import Basis, Pauli
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Position3D

if TYPE_CHECKING:
    from pyzx.graph.graph_s import GraphS
    from pyzx.pauliweb import PauliWeb

    from tqec.computation._correlation import _CorrelationSurface
    from tqec.computation.block_graph import BlockGraph


class ZXNode(NamedTuple):
    """Represent a node in the ZX graph spanned by the correlation surface.

    Correlation surface is represented by a set of edges in the ZX graph. Each edge
    consists of two nodes. The node id refers to the vertex index in the graph
    and the basis represents the Pauli operator on the half edge incident to the node.

    Attributes:
        id: The index of the vertex in the graph.
        basis: The Pauli operator on the half edge incident to the node.

    """

    id: int
    basis: Basis


class ZXEdge(NamedTuple):
    """Represent an edge in the ZX graph spanned by the correlation surface.

    Correlation surface is represented by a set of edges in the ZX graph. Each edge
    consists of two nodes. The edge can be decomposed into two half edges, each
    incident to a node. The half edges are labeled with the Pauli operator.
    If the edge has no hadamard effect, the Pauli operators on the two half edges
    should be the same. Otherwise, the Pauli operators should be flipped.

    Note that the Pauli operators are represented by `Basis` enum, which only
    contains `X` and `Z` operators. The `Y` operator is represented by two
    edges with `X` and `Z` operators respectively.

    """

    u: ZXNode
    v: ZXNode

    @staticmethod
    def sorted(u: ZXNode, v: ZXNode) -> ZXEdge:
        """Create a ZXEdge with nodes sorted."""
        return ZXEdge(*tuple(sorted([u, v])))

    @property
    def is_self_loop(self) -> bool:
        """Whether the edge is a self-loop edge.

        By definition, a self-loop edge represents a correlation surface within a single node. This
        is an edge case where the ZX graph only contains a single node.

        """
        return self.u.id == self.v.id

    @property
    def has_hadamard(self) -> bool:
        """Whether the edge has a hadamard effect."""
        return self.u.basis is not self.v.basis

    def get_basis(self, vertex_id: int) -> Basis:
        """Get the basis of the half edge incident to the given vertex."""
        match vertex_id:
            case self.u.id:
                return self.u.basis
            case self.v.id:
                return self.v.basis
            case _:
                raise TQECError(f"Vertex {vertex_id} is not incident to the edge {self}.")


@dataclass(frozen=True)
class CorrelationSurface:
    """Represent a set of measurements whose values determine the parity of the logical operators.

    A correlation surface in a computation is a set of measurements whose values determine the
    parity of the logical operators at the inputs and outputs associated with the surface.

    Note:
        We use the term "correlation surface", "pauli web" and "observable" interchangeably in
        the library.

    Here we represent the correlation surface by the set of edges it spans in the ZX graph.
    The insight is that the spiders pose parity constraints on the operators supported on
    the incident edges. The flow of the logical operators through the ZX graph, respecting
    the parity constraints, forms the correlation between the inputs and outputs. However,
    the sign of measurement outcomes is neglected in this representation. And we need to
    recover the measurements when instantiating an explicit logical observable from the
    correlation surface.

    Each edge establishes a correlation between the logical operators at the two ends of
    the edge. For example, an edge connecting two Z nodes represents the correlation
    between the logical Z operators at the two nodes.

    Attributes:
        span: A set of ``ZXEdge`` representing the span of the correlation surface.

    """

    span: frozenset[ZXEdge]

    @cached_property
    def _graph_view(self) -> tuple[dict[int, dict[int, list[ZXEdge]]], dict[int, set[Basis]]]:
        """Internal index mapping vertex IDs to active bases and incident edges."""
        edges, bases = {}, {}
        for edge in self.span:
            u, v = edge.u.id, edge.v.id
            edges.setdefault(u, {}).setdefault(v, []).append(edge)
            edges.setdefault(v, {}).setdefault(u, []).append(edge)
            bases.setdefault(u, set()).add(edge.u.basis)
            bases.setdefault(v, set()).add(edge.v.basis)
        return edges, bases

    def bases_at(self, v: int) -> set[Basis]:
        """Get the bases of the surfaces present at the vertex."""
        return self._graph_view[1].get(v, set())

    def to_pauli_web(self, g: GraphS) -> PauliWeb[int, tuple[int, int]]:
        """Convert the correlation surface to a Pauli web.

        Args:
            g: The ZX graph the correlation surface is based on.

        Returns:
            A `PauliWeb` representation of the correlation surface.

        """
        # Avoid pulling pyzx when importing that module.
        from tqec.interop.pyzx.correlation import correlation_surface_to_pauli_web  # noqa: PLC0415

        return correlation_surface_to_pauli_web(self, g)

    @staticmethod
    def from_pauli_web(pauli_web: PauliWeb[int, tuple[int, int]]) -> CorrelationSurface:
        """Create a correlation surface from a Pauli web."""
        # Avoid pulling pyzx when importing that module.
        from tqec.interop.pyzx.correlation import pauli_web_to_correlation_surface  # noqa: PLC0415

        return pauli_web_to_correlation_surface(pauli_web)

    @property
    def is_single_node(self) -> bool:
        """Whether the correlation surface contains only a single node.

        This is an edge case where the ZX graph only contains a single node. The span of the
        correlation surface is a self-loop edge at the node.

        """
        return len(self.span) == 1 and next(iter(self.span)).is_self_loop

    @property
    def span_vertices(self) -> set[int]:
        """Return the set of vertices in the correlation surface."""
        return set(self._graph_view[0].keys())

    def edges_at(self, v: int) -> set[ZXEdge]:
        """Return the set of edges incident to the vertex in the correlation surface."""
        return set(chain.from_iterable(self._graph_view[0].get(v, {}).values()))

    def external_stabilizer(self, io_ports: list[int]) -> str:
        """Get the Pauli operator supported on the given input/output ports.

        Args:
            io_ports: The list of input/output ports to consider.

        Returns:
            The Pauli operator supported on the given ports.

        """
        return "".join(str(Pauli.from_basis_set(self.bases_at(port))) for port in io_ports)

    def external_stabilizer_on_graph(self, graph: BlockGraph) -> str:
        """Get the external stabilizer of the correlation surface on the graph.

        If the provided graph is an open graph, the external stabilizer is the Pauli
        operator supported on the input/output ports of the graph. Otherwise, the
        external stabilizer is the Pauli operator supported on the leaf nodes of the
        graph.

        Args:
            graph: The block graph to consider.

        Returns:
            The Pauli operator that is the external stabilizer of the correlation surface.

        """
        supports: list[Position3D]
        if graph.is_open:
            port_labels = graph.ordered_ports
            supports = [graph.ports[p] for p in port_labels]
        else:
            supports = [cube.position for cube in graph.leaf_cubes]
        zx = graph.to_zx_graph()
        p2v = zx.p2v
        zx_ports = [p2v[p] for p in supports]
        return self.external_stabilizer(zx_ports)

    @cached_property
    def area(self) -> int:
        """Return the area of the correlation surface.

        The area of the correlation surface is the number of nodes it spans. A X node and a Z node
        with the same id are counted as two nodes.

        """
        span_nodes = {node for edge in self.span for node in edge}
        return len(span_nodes)

    def __xor__(self, other: CorrelationSurface) -> CorrelationSurface:
        return CorrelationSurface(self.span.symmetric_difference(other.span))

    def _to_mutable_graph_representation(self, zx_graph: GraphS) -> _CorrelationSurface:
        """Convert to the internal mutable representation."""
        # Avoid pulling pyzx when importing that module.
        from tqec.computation._correlation import _CorrelationSurface  # noqa: PLC0415
        from tqec.interop.pyzx.utils import is_hadamard  # noqa: PLC0415

        surface = _CorrelationSurface()
        for u, edges in self._graph_view[0].items():
            for v, edge in edges.items():
                surface._add_pauli_to_edge(
                    (u, v),
                    reduce(xor, (e.get_basis(u).to_pauli() for e in edge)),
                    is_hadamard(zx_graph, (u, v)),
                )
        for u, v in zx_graph.edges():
            if u not in surface or v not in surface[u]:
                surface._add_pauli_to_edge((u, v), Pauli.I, False)
        return surface


def find_correlation_surfaces(
    g: GraphS,
    vertex_ordering: Sequence[set[int]] | None = None,
    parallel: bool = False,
) -> list[CorrelationSurface]:
    """Find the correlation surfaces in a ZX graph.

    The function explores how can the X/Z logical observable move through the graph to form a
    correlation surface with the following steps:

    1. Identify connected components in the graph.
    2. For each connected component, run the following algorithm from the smallest leaf node to find
       a generating set of correlation surfaces assuming all ports are open:
         a. Explore the graph node-by-node while keeping a generating set of correlation surfaces
            for the subgraph explored.
         b. Generate valid correlation surfaces given the newly explored node.
         c. If there is a loop, recover valid correlation surfaces from invalid ones.
         d. Prune redundant ones to keep the generating set minimal.
         e. Repeat from step (a) until all nodes are explored.
    3. Reform the generators so that they satisfy the closed ports (non-BOUNDARY leaf nodes).
    4. Reform the generators so that the number of Y-terminating correlation surfaces is minimized.
    5. Combine the generators from all connected components.

    The rules for generating valid correlation surfaces at each node are as follows. For a node of
    basis B in {X, Z}:
    - *broadcast rule:* All or none of the incident edges supports the opposite of B.
    - *passthrough rule:* An even number of incident edges supports B.

    Leaf nodes can additionally be of both or none of the X and Z bases and also need to follow the
      above rules. Specifically:
    - For an X/Z type leaf node, it can only support the logical observable with the opposite type.
    - For a Y type leaf node, it can only support the Y logical observable, i.e. the presence of
      both X and Z logical observable.
    - For the BOUNDARY node, it can support any type of logical observable.

    Args:
        g: The ZX graph to find the correlation surfaces.
        vertex_ordering: A reserved argument for an unfinished feature. Should not be used at this
            moment.
        parallel: Whether to use multiprocessing to speed up the computation. Only applies to
            embarrassingly parallel parts of the algorithm. Default is `False`.

    Returns:
        A list of `CorrelationSurface` in the graph.

    """
    # Needs to be imported here to avoid pulling pyzx when importing this module.
    from tqec.computation._correlation import (  # noqa: PLC0415
        _check_spiders_are_supported,
        _find_correlation_surfaces_with_vertex_ordering,
    )
    from tqec.interop.pyzx.utils import zx_to_basis  # noqa: PLC0415

    if vertex_ordering is not None:
        raise NotImplementedError(
            "The `vertex_ordering` argument is reserved for an unfinished feature and should not"
            " be used at this moment."
        )
    _check_spiders_are_supported(g)
    # Edge case: single node graph
    if g.num_vertices() == 1:
        v = next(iter(g.vertices()))
        node = ZXNode(v, zx_to_basis(g, v).flipped())
        return [CorrelationSurface(frozenset({ZXEdge(node, node)}))]

    leaves = {v for v in g.vertices() if g.vertex_degree(v) == 1}
    if not leaves:
        raise TQECError(
            "The graph must contain at least one leaf node to find correlation surfaces."
        )

    # sort the correlation surfaces by area
    return sorted(
        (
            cs._to_immutable_public_representation(g)
            for cs in _find_correlation_surfaces_with_vertex_ordering(g, vertex_ordering, parallel)
        ),
        key=lambda x: sorted(x.span),
    )


def reduce_observables_to_minimal_generators(
    stabilizers_to_surfaces: dict[str, CorrelationSurface],
    hint_num_generators: int | None = None,
) -> dict[str, CorrelationSurface]:
    """Reduce a set of observables to generators with the smallest correlation surface area.

    Args:
        stabilizers_to_surfaces: The mapping from the stabilizer to the correlation surface.
        hint_num_generators: The hint number of generators to find. If provided,
            the function will stop after finding the specified number of generators.
            Otherwise, the function will iterate through all the stabilizers.

    Returns:
        A mapping from the generators' stabilizers to the correlation surfaces.

    """
    if not stabilizers_to_surfaces:
        return {}
    # Sort the stabilizers by its corresponding correlation surface's area to
    # find the generators with the smallest area
    stabs_ordered_by_area = sorted(
        stabilizers_to_surfaces.keys(),
        key=lambda s: (stabilizers_to_surfaces[s].area, s),
    )
    # find a complete set of generators, starting from the smallest area
    generators: list[str] = stabs_ordered_by_area[:1]
    generators_stim: list[stim.PauliString] = [stim.PauliString(generators[0])]
    for stabilizer in stabs_ordered_by_area[1:]:
        pauli_string = stim.PauliString(stabilizer)
        if not _can_be_generated_by(pauli_string, generators_stim):
            generators.append(stabilizer)
            generators_stim.append(pauli_string)
        if hint_num_generators is not None and len(generators) == hint_num_generators:
            break
    return {g: stabilizers_to_surfaces[g] for g in generators}


def _can_be_generated_by(
    pauli_string: stim.PauliString,
    basis: list[stim.PauliString],
) -> bool:
    """Check if the given Pauli string can be generated by the given basis.

    The following procedure is used to check if a Pauli string can be generated by
    a given basis of stabilizers:

    1. ``stim.Tableau.from_stabilizers`` constructs a tableau, which represents
       a Clifford circuit mapping the Z Pauli on the nth input to the nth
       stabilizer specified in the argument at the output. That means that the
       inverse tableau maps the nth stabilizer to Z on the nth output.
    2. Apply the inverse tableau to the Pauli string. If the Pauli string is a
       product of the stabilizers, it can be decomposed into them. As a result,
       after applying the tableau, the result will have `Z` operators only on
       the specific outputs. If not, the Pauli string cannot be generated by the
       given stabilizers.

    """
    if not basis:
        return False
    tableau = stim.Tableau.from_stabilizers(
        basis,
        allow_redundant=False,
        allow_underconstrained=True,
    )
    inv = tableau.inverse(unsigned=True)
    out_paulis = inv(pauli_string)
    # use `bool()` to avoid type error
    return bool(out_paulis[len(basis) :].weight == 0)



"""Implementation of different noise models for Stim.

Important note:
---------------

This file has been recovered from https://zenodo.org/records/7487893
and is under CC BY 4.0 (https://creativecommons.org/licenses/by/4.0/legalcode).
It is part of the code from the paper

    Gidney, C. (2022). Data for "Inplace Access to the Surface Code Y Basis".
    https://doi.org/10.5281/zenodo.7487893


Modifications to the original code:

1. Formatting with ruff.
2. Fixing typing issues and adapting a few imports to personal taste.
3. A minor adjustment to the measurement noise rules of ``uniform_depolarizing``
   and ``si1000`` noise models: the depolarizing error on the measured qubits
   after measurement is removed because, in our library, all measurements are
   immediately followed by resets. As a result, the depolarizing error has no
   effect.
4. Remove the ``depolarizing_two_body_measurement_noise`` noise model.
5. Changes to NoiseModel.noisy_circuit to respect TQEC convention with TICKs and
   REPEAT blocks (not before the block, the first instruction in the repeated
   inner block, and after the block).
6. Re-phrase the docstrings and error messages slightly.

"""

from collections import Counter, defaultdict
from collections.abc import Iterator, Set

import stim

CLIFFORD_1Q = "C1"
CLIFFORD_2Q = "C2"
ANNOTATION = "info"
MPP = "MPP"
MEASURE_RESET_1Q = "MR1"
JUST_MEASURE_1Q = "M1"
JUST_RESET_1Q = "R1"
NOISE = "!?"

OP_TYPES = {
    "I": CLIFFORD_1Q,
    "X": CLIFFORD_1Q,
    "Y": CLIFFORD_1Q,
    "Z": CLIFFORD_1Q,
    "C_XYZ": CLIFFORD_1Q,
    "C_ZYX": CLIFFORD_1Q,
    "H": CLIFFORD_1Q,
    "H_XY": CLIFFORD_1Q,
    "H_XZ": CLIFFORD_1Q,
    "H_YZ": CLIFFORD_1Q,
    "S": CLIFFORD_1Q,
    "SQRT_X": CLIFFORD_1Q,
    "SQRT_X_DAG": CLIFFORD_1Q,
    "SQRT_Y": CLIFFORD_1Q,
    "SQRT_Y_DAG": CLIFFORD_1Q,
    "SQRT_Z": CLIFFORD_1Q,
    "SQRT_Z_DAG": CLIFFORD_1Q,
    "S_DAG": CLIFFORD_1Q,
    "CNOT": CLIFFORD_2Q,
    "CX": CLIFFORD_2Q,
    "CY": CLIFFORD_2Q,
    "CZ": CLIFFORD_2Q,
    "ISWAP": CLIFFORD_2Q,
    "ISWAP_DAG": CLIFFORD_2Q,
    "SQRT_XX": CLIFFORD_2Q,
    "SQRT_XX_DAG": CLIFFORD_2Q,
    "SQRT_YY": CLIFFORD_2Q,
    "SQRT_YY_DAG": CLIFFORD_2Q,
    "SQRT_ZZ": CLIFFORD_2Q,
    "SQRT_ZZ_DAG": CLIFFORD_2Q,
    "SWAP": CLIFFORD_2Q,
    "XCX": CLIFFORD_2Q,
    "XCY": CLIFFORD_2Q,
    "XCZ": CLIFFORD_2Q,
    "YCX": CLIFFORD_2Q,
    "YCY": CLIFFORD_2Q,
    "YCZ": CLIFFORD_2Q,
    "ZCX": CLIFFORD_2Q,
    "ZCY": CLIFFORD_2Q,
    "ZCZ": CLIFFORD_2Q,
    "MPP": MPP,
    "MR": MEASURE_RESET_1Q,
    "MRX": MEASURE_RESET_1Q,
    "MRY": MEASURE_RESET_1Q,
    "MRZ": MEASURE_RESET_1Q,
    "M": JUST_MEASURE_1Q,
    "MX": JUST_MEASURE_1Q,
    "MY": JUST_MEASURE_1Q,
    "MZ": JUST_MEASURE_1Q,
    "R": JUST_RESET_1Q,
    "RX": JUST_RESET_1Q,
    "RY": JUST_RESET_1Q,
    "RZ": JUST_RESET_1Q,
    "DETECTOR": ANNOTATION,
    "OBSERVABLE_INCLUDE": ANNOTATION,
    "QUBIT_COORDS": ANNOTATION,
    "SHIFT_COORDS": ANNOTATION,
    "TICK": ANNOTATION,
    "E": ANNOTATION,
    "DEPOLARIZE1": NOISE,
    "DEPOLARIZE2": NOISE,
    "PAULI_CHANNEL_1": NOISE,
    "PAULI_CHANNEL_2": NOISE,
    "X_ERROR": NOISE,
    "Y_ERROR": NOISE,
    "Z_ERROR": NOISE,
    # Not supported.
    # 'CORRELATED_ERROR': NOISE,
    # 'E': NOISE,
    # 'ELSE_CORRELATED_ERROR',
}
OP_MEASURE_BASES = {
    "M": "Z",
    "MX": "X",
    "MY": "Y",
    "MZ": "Z",
    "MPP": "",
}
COLLAPSING_OPS = {
    op for op, t in OP_TYPES.items() if t in {JUST_RESET_1Q, JUST_MEASURE_1Q, MPP, MEASURE_RESET_1Q}
}


class NoiseRule:
    """Describes how to add noise to an operation."""

    def __init__(self, *, after: dict[str, float], flip_result: float = 0):
        """Describe how to add noise to an operation.

        Args:
            after: A dictionary mapping noise rule names to their probability argument.
                For example, {"DEPOLARIZE2": 0.01, "X_ERROR": 0.02} will add two qubit
                depolarization with parameter 0.01 and also add 2% bit flip noise. These
                noise channels occur after all other operations in the moment and are applied
                to the same targets as the relevant operation.
            flip_result: The probability that a measurement result should be reported incorrectly.
                Only valid when applied to operations that produce measurement results.

        """
        if not (0 <= flip_result <= 1):
            raise ValueError(f"not (0 <= {flip_result=} <= 1)")
        for k, p in after.items():
            if OP_TYPES[k] != NOISE:
                raise ValueError(f"not a noise channel: {k} from {after=}")
            if not (0 <= p <= 1):
                raise ValueError(f"not (0 <= {p} <= 1) from {after=}")
        self.after = after
        self.flip_result = flip_result

    def _append_noisy_version_of(
        self,
        *,
        split_op: stim.CircuitInstruction,
        out_during_moment: stim.Circuit,
        after_moments: defaultdict[tuple[str, float], stim.Circuit],
        immune_qubits: Set[int],
    ) -> None:
        targets = split_op.targets_copy()
        if immune_qubits and any(
            (t.is_qubit_target or t.is_x_target or t.is_y_target or t.is_z_target)
            and t.value in immune_qubits
            for t in targets
        ):
            out_during_moment.append(split_op)
            return

        args = split_op.gate_args_copy()
        if self.flip_result:
            t = OP_TYPES[split_op.name]
            assert t in {MPP, JUST_MEASURE_1Q, MEASURE_RESET_1Q}
            assert len(args) == 0
            args = [self.flip_result]

        out_during_moment.append(split_op.name, targets, args)
        raw_targets = [t.value for t in targets if not t.is_combiner]
        for op_name, arg in self.after.items():
            after_moments[(op_name, arg)].append(op_name, raw_targets, arg)


class NoiseModel:
    def __init__(
        self,
        idle_depolarization: float,
        additional_depolarization_waiting_for_m_or_r: float = 0,
        gate_rules: dict[str, NoiseRule] | None = None,
        measure_rules: dict[str, NoiseRule] | None = None,
        any_clifford_1q_rule: NoiseRule | None = None,
        any_clifford_2q_rule: NoiseRule | None = None,
    ):
        """Represent a noise model that can be applied to a ``stim.Circuit``."""
        self.idle_depolarization = idle_depolarization
        self.additional_depolarization_waiting_for_m_or_r = (
            additional_depolarization_waiting_for_m_or_r
        )
        self.gate_rules = gate_rules
        self.measure_rules = measure_rules
        self.any_clifford_1q_rule = any_clifford_1q_rule
        self.any_clifford_2q_rule = any_clifford_2q_rule

    @staticmethod
    def si1000(p: float) -> "NoiseModel":
        """Superconducting inspired noise.

        As defined in "A Fault-Tolerant Honeycomb Memory":
        https://arxiv.org/abs/2108.10457

        Small tweak from the paper: The measurement result is probabilistically flipped instead of
        the input qubit.

        Note that this noise model only allows Z basis measurements and resets. Currently,
        tqec does not automatically compile to compatible physical circuits.
        """
        return NoiseModel(
            idle_depolarization=p / 10,
            additional_depolarization_waiting_for_m_or_r=2 * p,
            any_clifford_1q_rule=NoiseRule(after={"DEPOLARIZE1": p / 10}),
            any_clifford_2q_rule=NoiseRule(after={"DEPOLARIZE2": p}),
            measure_rules={
                "Z": NoiseRule(after={}, flip_result=p * 5),
            },
            gate_rules={
                "R": NoiseRule(after={"X_ERROR": p * 2}),
            },
        )

    @staticmethod
    def uniform_depolarizing(p: float) -> "NoiseModel":
        """Near-standard circuit depolarizing noise.

        Everything has the same parameter ``p``. Single-qubit Clifford gates get single-qubit
        depolarization. Two-qubit Clifford gates get single-qubit depolarization. Dissipative gates
        have their result probabilistically bit-flipped (or phase-flipped if appropriate).

        Non-demolition measurement is treated a bit unusually in that it is the result that is
        flipped instead of the input qubit.

        """
        return NoiseModel(
            idle_depolarization=p,
            any_clifford_1q_rule=NoiseRule(after={"DEPOLARIZE1": p}),
            any_clifford_2q_rule=NoiseRule(after={"DEPOLARIZE2": p}),
            measure_rules={
                "X": NoiseRule(after={}, flip_result=p),
                "Y": NoiseRule(after={}, flip_result=p),
                "Z": NoiseRule(after={}, flip_result=p),
                "XX": NoiseRule(after={}, flip_result=p),
                "YY": NoiseRule(after={}, flip_result=p),
                "ZZ": NoiseRule(after={}, flip_result=p),
            },
            gate_rules={
                "RX": NoiseRule(after={"Z_ERROR": p}),
                "RY": NoiseRule(after={"X_ERROR": p}),
                "R": NoiseRule(after={"X_ERROR": p}),
            },
        )

    def _noise_rule_for_split_operation(
        self, *, split_op: stim.CircuitInstruction
    ) -> NoiseRule | None:
        if occurs_in_classical_control_system(split_op):
            return None

        if self.gate_rules is not None:
            rule = self.gate_rules.get(split_op.name)
            if rule is not None:
                return rule

        t = OP_TYPES[split_op.name]

        if self.any_clifford_1q_rule is not None and t == CLIFFORD_1Q:
            return self.any_clifford_1q_rule
        if self.any_clifford_2q_rule is not None and t == CLIFFORD_2Q:
            return self.any_clifford_2q_rule
        if self.measure_rules is not None and t in (MPP, JUST_MEASURE_1Q):
            measure_basis = _measure_basis(split_op=split_op)
            assert measure_basis is not None
            rule = self.measure_rules.get(measure_basis)
            if rule is not None:
                return rule

        raise ValueError(
            f"No noise (or lack of noise) specified for {split_op=}. Please make "
            "sure the noise model covers all operations in the compiled circuit."
        )

    def _append_idle_error(
        self,
        *,
        moment_split_ops: list[stim.CircuitInstruction],
        out: stim.Circuit,
        system_qubits: Set[int],
        immune_qubits: Set[int],
    ) -> None:
        collapse_qubits: list[int] = []
        clifford_qubits: list[int] = []
        for split_op in moment_split_ops:
            if occurs_in_classical_control_system(split_op):
                continue
            if split_op.name in COLLAPSING_OPS:
                qubits_out = collapse_qubits
            else:
                qubits_out = clifford_qubits
            for target in split_op.targets_copy():
                if not target.is_combiner:
                    qubits_out.append(target.value)

        # Safety check for operation collisions.
        usage_counts = Counter(collapse_qubits + clifford_qubits)
        qubits_used_multiple_times = {q for q, c in usage_counts.items() if c != 1}
        if qubits_used_multiple_times:
            moment = stim.Circuit()
            for op in moment_split_ops:
                moment.append(op)
            raise ValueError(
                f"Qubits were operated on multiple times without a TICK in between:\n"
                f"multiple uses: {sorted(qubits_used_multiple_times)}\n"
                f"moment:\n"
                f"{moment}"
            )

        collapse_qubits_set = set(collapse_qubits)
        clifford_qubits_set = set(clifford_qubits)
        idle = sorted(system_qubits - collapse_qubits_set - clifford_qubits_set - immune_qubits)
        if idle and self.idle_depolarization:
            out.append("DEPOLARIZE1", idle, self.idle_depolarization)

        waiting_for_mr = sorted(system_qubits - collapse_qubits_set - immune_qubits)
        if (
            collapse_qubits_set
            and waiting_for_mr
            and self.additional_depolarization_waiting_for_m_or_r
        ):
            out.append("DEPOLARIZE1", idle, self.additional_depolarization_waiting_for_m_or_r)

    def _append_noisy_moment(
        self,
        *,
        moment_split_ops: list[stim.CircuitInstruction],
        out: stim.Circuit,
        system_qubits: Set[int],
        immune_qubits: Set[int],
    ) -> None:
        after: defaultdict[tuple[str, float], stim.Circuit] = defaultdict(stim.Circuit)
        for split_op in moment_split_ops:
            rule = self._noise_rule_for_split_operation(split_op=split_op)
            if rule is None:
                out.append(split_op)
            else:
                rule._append_noisy_version_of(
                    split_op=split_op,
                    out_during_moment=out,
                    after_moments=after,
                    immune_qubits=immune_qubits,
                )
        for k in sorted(after.keys()):
            out += after[k]

        self._append_idle_error(
            moment_split_ops=moment_split_ops,
            out=out,
            system_qubits=system_qubits,
            immune_qubits=immune_qubits,
        )

    def noisy_circuit(
        self,
        circuit: stim.Circuit,
        *,
        system_qubits: set[int] | None = None,
        immune_qubits: set[int] | None = None,
    ) -> stim.Circuit:
        """Return a noisy version of the given circuit, by applying the receiving noise model.

        Args:
            circuit: The circuit to layer noise over.
            system_qubits: All qubits used by the circuit. These are the qubits eligible for idling
                noise.
            immune_qubits: Qubits to not apply noise to, even if they are operated on.

        Returns:
            The noisy version of the circuit.

        """
        if system_qubits is None:
            system_qubits = set(range(circuit.num_qubits))
        if immune_qubits is None:
            immune_qubits = set()

        result = stim.Circuit()
        for moment_split_ops in _iter_split_op_moments(circuit, immune_qubits=immune_qubits):
            if not result:
                pass
            elif isinstance(moment_split_ops, stim.CircuitRepeatBlock):
                pass
            elif isinstance(result[-1], stim.CircuitRepeatBlock):
                pass
            else:
                result.append("TICK", [], [])
            if isinstance(moment_split_ops, stim.CircuitRepeatBlock):
                noisy_body = self.noisy_circuit(
                    moment_split_ops.body_copy(),
                    system_qubits=system_qubits,
                    immune_qubits=immune_qubits,
                )
                result.append(
                    stim.CircuitRepeatBlock(
                        repeat_count=moment_split_ops.repeat_count, body=noisy_body
                    )
                )
            else:
                self._append_noisy_moment(
                    moment_split_ops=moment_split_ops,
                    out=result,
                    system_qubits=system_qubits,
                    immune_qubits=immune_qubits,
                )

        return result


def occurs_in_classical_control_system(op: stim.CircuitInstruction) -> bool:
    """Determine if an operation is an annotation or a classical control system update."""
    t = OP_TYPES[op.name]
    if t == ANNOTATION:
        return True
    if t == CLIFFORD_2Q:
        targets = op.targets_copy()
        for k in range(0, len(targets), 2):
            a = targets[k]
            b = targets[k + 1]
            classical_0 = a.is_measurement_record_target or a.is_sweep_bit_target
            classical_1 = b.is_measurement_record_target or b.is_sweep_bit_target
            if not (classical_0 or classical_1):
                return False
        return True
    return False


def _split_targets_if_needed(
    op: stim.CircuitInstruction, immune_qubits: Set[int]
) -> Iterator[stim.CircuitInstruction]:
    """Split operations into pieces as needed.

    This function splits operations, for example ``MPP`` into each product, classical control away
    from quantum ops, ...

    """
    t = OP_TYPES[op.name]
    if t == CLIFFORD_2Q:
        yield from _split_targets_if_needed_clifford_2q(op, immune_qubits)
    elif t == MPP:
        yield from _split_targets_if_needed_m_basis(op, immune_qubits)
    elif t in [NOISE, ANNOTATION]:
        yield op
    else:
        yield from _split_targets_if_needed_clifford_1q(op, immune_qubits)


def _split_targets_if_needed_clifford_1q(
    op: stim.CircuitInstruction, immune_qubits: Set[int]
) -> Iterator[stim.CircuitInstruction]:
    if immune_qubits:
        args = op.gate_args_copy()
        for t in op.targets_copy():
            yield stim.CircuitInstruction(op.name, [t], args)
    else:
        yield op


def _split_targets_if_needed_clifford_2q(
    op: stim.CircuitInstruction, immune_qubits: Set[int]
) -> Iterator[stim.CircuitInstruction]:
    """Split classical control system operations away from quantum operations."""
    assert OP_TYPES[op.name] == CLIFFORD_2Q
    targets = op.targets_copy()
    if immune_qubits or any(t.is_measurement_record_target for t in targets):
        args = op.gate_args_copy()
        for k in range(0, len(targets), 2):
            yield stim.CircuitInstruction(op.name, targets[k : k + 2], args)
    else:
        yield op


def _split_targets_if_needed_m_basis(
    op: stim.CircuitInstruction, immune_qubits: Set[int]
) -> Iterator[stim.CircuitInstruction]:
    """Split an MPP operation into one operation for each Pauli product it measures."""
    targets = op.targets_copy()
    args = op.gate_args_copy()
    k = 0
    start = k
    while k < len(targets):
        if k + 1 == len(targets) or not targets[k + 1].is_combiner:
            yield stim.CircuitInstruction(op.name, targets[start : k + 1], args)
            k += 1
            start = k
        else:
            k += 2
    assert k == len(targets)


def _iter_split_op_moments(
    circuit: stim.Circuit, *, immune_qubits: Set[int]
) -> Iterator[stim.CircuitRepeatBlock | list[stim.CircuitInstruction]]:
    """Split a circuit into moments and some operations into pieces.

    Classical control system operations like ``CX rec[-1] 0`` are split from quantum operations
    like ``CX 1 0``.

    MPP operations are split into one operation per Pauli product.

    Yields:
        Lists of operations corresponding to one moment in the circuit, with any problematic
        operations like MPPs split into pieces.

        (A moment is the time between two TICKs.)

    """
    cur_moment: list[stim.CircuitInstruction] = []

    for op in circuit:
        if isinstance(op, stim.CircuitRepeatBlock):
            if cur_moment:
                yield cur_moment
                cur_moment = []
            yield op
        elif op.name == "TICK":
            yield cur_moment
            cur_moment = []
        else:
            cur_moment.extend(_split_targets_if_needed(op, immune_qubits=immune_qubits))
    if cur_moment:
        yield cur_moment


def _measure_basis(*, split_op: stim.CircuitInstruction) -> str | None:
    """Convert an operation into a string describing the Pauli product basis it measures.

    Returns:
        None: This is not a measurement (or not *just* a measurement).
        str: Pauli product string that the operation measures (e.g. "XX" or "Y").

    """
    result = OP_MEASURE_BASES.get(split_op.name)
    if result == "":
        targets = split_op.targets_copy()
        for k in range(0, len(targets), 2):
            t = targets[k]
            if t.is_x_target:
                result += "X"
            elif t.is_y_target:
                result += "Y"
            elif t.is_z_target:
                result += "Z"
            else:
                raise NotImplementedError(f"{targets=}")
    return result



"""Collection of pre-built logical computations.

This module contains functions that build :py:class:`~tqec.computation.block_graph.BlockGraph`
instances representing the logical computations, including:

- :mod:`.memory`: memory experiment
- :mod:`.stability`: stability experiment
- :mod:`.cnot`: logical CNOT gate
- :mod:`.cz`: logical CZ gate
- :mod:`.move_rotation`: rotate spatial boundaries by moving the logical qubit in spacetime
- :mod:`.three_cnots`: three logical CNOT gates compressed in spacetime
- :mod:`.steane_encoding`: Steane encoding circuit compressed in spacetime

"""

from .cnot import cnot as cnot
from .cz import cz as cz
from .memory import memory as memory
from .move_rotation import move_rotation as move_rotation
from .stability import stability as stability
from .steane_encoding import steane_encoding as steane_encoding
from .three_cnots import three_cnots as three_cnots



"""Defines :meth:`.generate_circuit`, one of the core method of the :mod:`tqec` package.

This module defines two of core methods of the :mod:`tqec` package:

- :meth:`generate_circuit` that is the most convenient method for external users.
- :meth:`generate_circuit_from_instantiation` that gives more freedom to the
  user (and is used internally by :meth:`generate_circuit`) at the expense of
  often less convenient inputs.

Both of these methods are used to generate the ``stim.Circuit`` instance
representing **one** time slice (often equivalent to "one QEC round"). Users are
expected to call these methods several times and concatenate the output
``stim.Circuit`` instances in time to obtain a full QEC implementation.

Note that these methods do not work with ``REPEAT`` instructions.

"""

from __future__ import annotations

import numpy
import numpy.typing as npt

from tqec.circuit.schedule import (
    ScheduledCircuit,
    merge_scheduled_circuits,
    relabel_circuits_qubit_indices,
)
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.base import Template
from tqec.utils.array import to2dlist
from tqec.utils.position import Shift2D


def generate_circuit(template: Template, k: int, plaquettes: Plaquettes) -> ScheduledCircuit:
    """Generate a quantum circuit from a template and its plaquettes.

    This is one of the core methods of the :mod:`tqec` package. It generates a
    quantum circuit from the description of the template that should be
    implemented as well as the plaquettes that should be used to instantiate the
    provided template.

    This function requires that a few pre-conditions on the inputs are met:

    1. the number of plaquettes provided should match the number of plaquettes
       required by the provided template.
    2. all the provided plaquettes should be implemented on
       :class:`~.qubit.GridQubit` instances **only**.

    If any of the above pre-conditions is not met, the inputs are considered
    invalid, in which case this function **might** raise an error.

    Args:
        template: spatial description of the quantum error correction experiment
            we want to implement.
        k: scaling parameter used to instantiate the provided ``template``.
        plaquettes: description of the computation that should happen at
            different time-slices of the quantum error correction experiment (or
            at least part of it).

    Returns:
        a :class:`~.schedule.circuit.ScheduledCircuit` instance implementing the
        (part of) quantum error correction experiment represented by the
        provided inputs.

    """
    # instantiate the template with the appropriate plaquette indices.
    # Index 0 is "no plaquette" by convention and should not be included here.
    _indices = list(range(1, template.expected_plaquettes_number + 1))
    template_plaquettes = template.instantiate(k, _indices)
    increments = template.get_increments()

    return generate_circuit_from_instantiation(template_plaquettes, plaquettes, increments)


def generate_circuit_from_instantiation(
    plaquette_array: npt.NDArray[numpy.int_],
    plaquettes: Plaquettes,
    increments: Shift2D,
) -> ScheduledCircuit:
    """Generate a quantum circuit from an array of plaquette indices and the associated plaquettes.

    This is one of the core methods of the :mod:`tqec` package. It generates a
    quantum circuit from a spatial description of where the plaquettes should be
    located as well as the actual plaquettes used.

    This function requires that a few pre-conditions on the inputs are met:

    1. the number of plaquettes provided should match the number of plaquettes
       required by the provided template.
    2. all the provided plaquettes should be implemented on
       :class:`~.qubit.GridQubit` instances **only**.

    If any of the above pre-conditions is not met, the inputs are considered
    invalid, in which case this function **might** raise an error.

    Args:
        plaquette_array: a 2-dimensional array of indices referencing
            :class:`~tqec.plaquette.plaquette.Plaquette` instances in the
            ``plaquettes`` argument.
        plaquettes: description of the computation that should happen at
            different time-slices of the quantum error correction experiment (or
            at least part of it).
        increments: the displacement between each plaquette origin.

    Returns:
        a :class:`~.schedule.circuit.ScheduledCircuit` instance implementing the
        (part of) quantum error correction experiment represented by the
        provided inputs.

    Raises:
        TQECError: if any index in ``plaquette_array`` is not correctly
            associated to a plaquette in ``plaquettes``.

    """
    # Collect all the used plaquette indices, removing 0 if present.
    indices = numpy.unique(plaquette_array)
    if indices[0] == 0:
        indices = indices[1:]

    # Plaquettes indices are starting at 1 in template_plaquettes. To avoid
    # offsets in the following code, we add an empty circuit at position 0.
    plaquette_circuits = {0: ScheduledCircuit.empty()} | {
        i: plaquettes[i].circuit for i in indices.tolist()
    }

    # Generate the ScheduledCircuit instances for each plaquette instantiation
    all_scheduled_circuits: list[ScheduledCircuit] = []
    additional_mergeable_instructions: set[str] = set()
    # The below line is not strictly needed, but makes type checkers happy with
    # type inference. See https://numpy.org/doc/stable/reference/typing.html#d-arrays
    # for more information on why this should be done.
    plaquette_array_list: list[list[int]] = to2dlist(plaquette_array)
    for row_index, line in enumerate(plaquette_array_list):
        for column_index, plaquette_index in enumerate(line):
            if plaquette_index != 0:
                # Computing the offset that should be applied to each qubits.
                plaquette = plaquettes[plaquette_index]
                qubit_offset = Shift2D(
                    plaquette.origin.x + column_index * increments.x,
                    plaquette.origin.y + row_index * increments.y,
                )
                # Warning: the variable `mapped_scheduled_circuit` shares with
                #          `plaquette_circuits[plaquette_index]` a reference to
                #          the circuit data-structure. This is not an issue here
                #          as we never attempt to mutate that circuit before
                #          calling `relabel_circuits_qubit_indices`, that
                #          explicitly returns a copy of its inputs, and do not
                #          mutate them either.
                scheduled_circuit = plaquette_circuits[plaquette_index]
                mapped_scheduled_circuit = scheduled_circuit.map_to_qubits(
                    lambda q: q + qubit_offset, inplace_qubit_map=False
                )
                all_scheduled_circuits.append(mapped_scheduled_circuit)
                additional_mergeable_instructions |= plaquette.mergeable_instructions

    # Merge everything, but first make sure that the circuits are compatible.
    # Note that relabel_circuits_qubit_indices guarantees in its documentation
    # that the input circuits are not mutated but rather copied. This allows us
    # to not deepcopy the circuits earlier in the function.
    all_scheduled_circuits, qubit_map = relabel_circuits_qubit_indices(all_scheduled_circuits)
    return merge_scheduled_circuits(
        all_scheduled_circuits, qubit_map, additional_mergeable_instructions
    )



from __future__ import annotations

from abc import ABC
from typing import Generic, cast

from typing_extensions import Self, TypeVar

from tqec.utils.exceptions import TQECError
from tqec.utils.position import BlockPosition2D, BlockPosition3D, SignedDirection3D


class LayoutPosition2D(ABC):
    def __init__(self, x: int, y: int) -> None:
        """Represent the local indexing used to represent both cubes and pipes.

        Args:
            x: first coordinate.
            y: second coordinate.

        """
        super().__init__()
        self._x = x
        self._y = y

    @staticmethod
    def from_block_position(pos: BlockPosition2D) -> LayoutCubePosition2D:
        """Get a layout position from a block position."""
        return LayoutCubePosition2D(2 * pos.x, 2 * pos.y)

    @staticmethod
    def from_pipe_position(
        pipe_position: tuple[BlockPosition2D, BlockPosition2D],
    ) -> LayoutPipePosition2D:
        """Get a layout pipe position from a pipe position."""
        u, v = sorted(pipe_position)
        assert u.is_neighbour(v)
        assert u < v
        return LayoutPipePosition2D(2 * u.x + (u.x != v.x), 2 * u.y + (u.y != v.y))

    def __hash__(self) -> int:
        return hash((self._x, self._y))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LayoutPosition2D) and self._x == other._x and self._y == other._y

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self._x},y={self._y})"

    def __add__(self, other: Self | tuple[int, int]) -> Self:
        if isinstance(other, tuple):
            x, y = cast(tuple[int, int], other)
            return self.__class__(self._x + x, self._y + y)
        return self.__class__(self._x + other._x, self._y + other._y)


class LayoutCubePosition2D(LayoutPosition2D):
    def __init__(self, x: int, y: int) -> None:
        """Represent the position of a cube on the grid.

        For the moment, only 2 entities have to appear on the grid: cubes and pipes.
        For that reason, we define cube positions (i.e., :class:`LayoutCubePosition2D`
        instances) to be on even coordinates and pipes positions to have one odd
        coordinates in the pipe dimension and even coordinates elsewhere.

        """
        if (x % 2 == 1) or (y % 2 == 1):
            clsname = self.__class__.__name__
            raise TQECError(f"{clsname} cannot contain any odd coordinate.")
        super().__init__(x, y)

    def to_block_position(self) -> BlockPosition2D:
        """Get the block position."""
        return BlockPosition2D(self._x // 2, self._y // 2)


class LayoutPipePosition2D(LayoutPosition2D):
    def __init__(self, x: int, y: int) -> None:
        """Represent the position of a cube on the grid.

        For the moment, only 2 entities have to appear on the grid: cubes and pipes.
        For that reason, we define cube positions (i.e., :class:`LayoutCubePosition2D`
        instances) to be on even coordinates and pipes positions to have one odd
        coordinates in the pipe dimension and even coordinates elsewhere.

        """
        if not ((x % 2 == 1) ^ (y % 2 == 1)):
            clsname = self.__class__.__name__
            raise TQECError(f"{clsname} should contain one odd and one even coordinate.")
        super().__init__(x, y)

    def to_pipe(self) -> tuple[BlockPosition2D, BlockPosition2D]:
        """Get the linked block positions."""
        if self._x % 2 == 1:
            return BlockPosition2D((self._x - 1) // 2, self._y // 2), BlockPosition2D(
                (self._x + 1) // 2, self._y // 2
            )
        return BlockPosition2D(self._x // 2, (self._y - 1) // 2), BlockPosition2D(
            self._x // 2, (self._y + 1) // 2
        )


T_co = TypeVar("T_co", bound=LayoutPosition2D, covariant=True, default=LayoutPosition2D)


class LayoutPosition3D(ABC, Generic[T_co]):
    def __init__(self, spatial_position: T_co, z: int) -> None:
        """Represent the local indexing used to represent both 3D cubes and pipes.

        This class simply wraps a :class:`LayoutPosition2D` instance with an integer-valued z
        coordinate.

        Because temporal pipes are "absorbed" in its neighbouring blocks, we do not
        have to represent them, hence the z coordinate does not need any kind of
        special treatment like the x and y coordinates.

        """
        super().__init__()
        self._spatial_position = spatial_position
        self._z = z

    @staticmethod
    def from_block_position(
        pos: BlockPosition3D,
    ) -> LayoutPosition3D[LayoutCubePosition2D]:
        """Get a layout position from a block position."""
        return LayoutPosition3D(LayoutCubePosition2D.from_block_position(pos.as_2d()), pos.z)

    @staticmethod
    def from_pipe_position(
        pipe_position: tuple[BlockPosition3D, BlockPosition3D],
    ) -> LayoutPosition3D[LayoutPipePosition2D]:
        """Get a layout pipe position from a pipe position."""
        u, v = sorted(pipe_position)
        assert u.is_neighbour(v)
        assert u < v
        return LayoutPosition3D(LayoutPosition2D.from_pipe_position((u.as_2d(), v.as_2d())), u.z)

    @staticmethod
    def from_block_and_signed_direction(
        pos: BlockPosition3D, dir: SignedDirection3D
    ) -> LayoutPosition3D[LayoutPipePosition2D]:
        """Get a 3-dimension pipe position from a block position and a direction.

        Args:
            pos: a block position connected by the returned pipe.
            dir: the direction in which we want the pipe to go out of ``pos``.

        Returns:
            the pipe position of the unique pipe linked to the block in ``pos`` in the provided
            ``dir``.

        """
        neighbour = pos.shift_in_direction(dir.direction, 1 if dir.towards_positive else -1)
        u, v = sorted((pos, neighbour))
        assert u.is_neighbour(v)
        return LayoutPosition3D(LayoutPosition2D.from_pipe_position((u.as_2d(), v.as_2d())), u.z)

    def __hash__(self) -> int:
        return hash((self._spatial_position, self._z))

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, LayoutPosition3D)
            and self._spatial_position == other._spatial_position
            and self._z == other._z
        )

    def as_2d(self) -> LayoutPosition2D:
        """Return a 2-dimensional position, ignoring the third dimension in ``self``."""
        return self._spatial_position

    @property
    def z(self) -> int:
        """Return the z-coordinate."""
        return self._z

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(spatial_position={self._spatial_position},z={self._z})"



"""Defines the :py:class:`~tqec.computation.cube.Cube` class."""

from __future__ import annotations

from dataclasses import astuple, dataclass
from typing import Any

from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, Position3D


@dataclass(frozen=True)
class ZXCube:
    """The kind of cubes consisting of only X or Z basis boundaries.

    Attributes:
        x: Looking at the cube along the x-axis, the basis of the walls observed.
        y: Looking at the cube along the y-axis, the basis of the walls observed.
        z: Looking at the cube along the z-axis, the basis of the walls observed.

    """

    x: Basis
    y: Basis
    z: Basis

    def __post_init__(self) -> None:
        if self.x == self.y == self.z:
            raise TQECError("The cube with the same basis along all axes is not allowed.")

    def as_tuple(self) -> tuple[Basis, Basis, Basis]:
        """Return a tuple of ``(self.x, self.y, self.z)``.

        Returns:
            A tuple of ``(self.x, self.y, self.z)``.

        """
        return (self.x, self.y, self.z)

    def __str__(self) -> str:
        return f"{self.x}{self.y}{self.z}"

    @staticmethod
    def all_kinds() -> list[ZXCube]:
        """Return all the allowed ``ZXCube`` instances.

        Returns:
            The list of all the allowed ``ZXCube`` instances.

        """
        return [ZXCube.from_str(s) for s in ["ZXZ", "XZZ", "ZXX", "XZX", "XXZ", "ZZX"]]

    @staticmethod
    def from_str(string: str) -> ZXCube:
        """Create a cube kind from the string representation.

        The string must be a 3-character string consisting of ``'X'`` or ``'Z'``,
        representing the basis of the walls along the x, y, and z axes.
        For example, a cube with left-right walls in the X basis, front-back walls in the Z basis,
        and top-bottom walls in the X basis can be constructed from the string ``'XZX'``.

        Args:
            string: A 3-character string consisting of ``'X'`` or ``'Z'``, representing
                the basis of the walls along the x, y, and z axes.

        Returns:
            The :py:class:`~tqec.computation.cube.ZXCube` instance constructed from
            the string representation.

        """
        return ZXCube(*map(Basis, string.upper()))

    @property
    def normal_basis(self) -> Basis:
        """Return the normal basis of the cube.

        Normal basis only appears once in the three bases of the cube. For example,
        the normal basis of the cube ``XZZ`` is ``X`` and the normal basis of the cube
        ``ZXX`` is ``Z``.

        """
        if sum(basis == Basis.Z for basis in astuple(self)) == 1:
            return Basis.Z
        return Basis.X

    @property
    def normal_direction(self) -> Direction3D:
        """Return the normal direction of the cube.

        Normal direction is the direction along which the normal basis appears.
        For example, the normal direction of the cube ``XZZ`` is ``Direction3D.X``
        and the normal direction of the cube ``XXZ`` is ``Direction3D.Z``.

        """
        return Direction3D(astuple(self).index(self.normal_basis))

    @property
    def is_spatial(self) -> bool:
        """Return whether a cube of this kind is a spatial cube.

        A spatial cube is a cube whose all spatial boundaries are in the same basis.
        There are only two possible spatial cubes: ``XXZ`` and ``ZZX``.

        """
        return self.x == self.y

    def get_basis_along(self, direction: Direction3D) -> Basis:
        """Get the basis of the walls along the given direction axis.

        Args:
            direction: The direction of the axis along which the basis is queried.

        Returns:
            The basis of the walls along the given direction axis.

        """
        return self.as_tuple()[direction.value]

    def with_basis_along(self, direction: Direction3D, basis: Basis) -> ZXCube:
        """Set the basis of the walls along the given direction axis and return a new instance.

        Args:
            direction: The direction of the axis along which the basis is set.
            basis: The basis to set along the given direction axis.

        Returns:
            The new :py:class:`~tqec.computation.cube.ZXCube` instance with the basis
            set along the given direction axis.

        """
        return ZXCube(
            *[basis if i == direction.value else b for i, b in enumerate(self.as_tuple())]
        )


class Port:
    """Cube kind representing the open ports in the block graph.

    The open ports correspond to the input/output of the computation represented by the block graph.
    They will have no effect on the functionality of the logical computation itself and should be
    invisible when visualizing the computation model.

    """

    def __str__(self) -> str:
        return "PORT"

    def __hash__(self) -> int:
        return hash(Port)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Port)


class YHalfCube:
    """Cube kind representing the Y-basis initialization/measurements."""

    def __str__(self) -> str:
        return "Y"

    def __hash__(self) -> int:
        return hash(YHalfCube)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, YHalfCube)


CubeKind = ZXCube | Port | YHalfCube
"""All the possible kinds of cubes."""


def cube_kind_from_string(s: str) -> CubeKind:
    """Create a cube kind from the string representation."""
    match s.upper():
        case "PORT" | "P":
            return Port()
        case "Y":
            return YHalfCube()
        case _:
            return ZXCube.from_str(s)


@dataclass(frozen=True)
class Cube:
    """A fundamental building block of the logical computation.

    A cube is a high-level abstraction of a block of quantum operations within a
    specific spacetime volume. These operations preserve or manipulate the quantum
    information encoded in the logical qubits.

    For example, a single ``ZXZ`` kind cube can represent a quantum memory experiment for
    a surface code patch in logical Z basis. The default circuit implementation of the
    cube will consist of transversal Z basis resets, syndrome extraction cycles, and finally
    the Z basis transversal measurements. The spatial location of the physical qubits in the
    code patch and the time when the operations are applied are specified by the spacetime
    position of the cube.

    Attributes:
        position: The position of the cube in the spacetime. The spatial coordinates
            determines which code patch the operations are applied to, and the time
            coordinate determines when the operations are applied.
        kind: The kind of the cube. It determines the basic logical operations represented
            by the cube.
        label: The label of the cube. It's mainly used for annotating the input/output
            ports of the block graph. If the cube is a port, the label must be non-empty
            and unique within the block graph. The label can be any string, but duplicate
            labels are not allowed. Default is an empty string.

    """

    position: Position3D
    kind: CubeKind
    label: str = ""

    def __post_init__(self) -> None:
        if self.is_port and not self.label:
            raise TQECError("A port cube must have a non-empty port label.")

    def __str__(self) -> str:
        return f"{self.kind}{self.position}"

    @property
    def is_zx_cube(self) -> bool:
        """Verify whether the cube is of kind :py:class:`~tqec.computation.cube.ZXCube`."""
        return isinstance(self.kind, ZXCube)

    @property
    def is_port(self) -> bool:
        """Verify whether the cube is of kind :py:class:`~tqec.computation.cube.Port`."""
        return isinstance(self.kind, Port)

    @property
    def is_y_cube(self) -> bool:
        """Verify whether the cube is of kind :py:class:`~tqec.computation.cube.YHalfCube`."""
        return isinstance(self.kind, YHalfCube)

    @property
    def is_spatial(self) -> bool:
        """Return whether the cube is a spatial cube.

        A spatial cube is a cube whose all spatial boundaries are in the same basis.
        There are only two possible spatial cubes: ``XXZ`` and ``ZZX``.

        """
        return isinstance(self.kind, ZXCube) and self.kind.is_spatial

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the cube."""
        return {
            "position": self.position.as_tuple(),
            "kind": str(self.kind),
            "label": self.label,
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Cube:
        """Create a cube from the dictionary representation.

        Args:
            data: The dictionary representation of the cube.

        Returns:
            The :py:class:`~tqec.computation.cube.Cube` instance created from the
            dictionary representation.

        """
        return Cube(
            position=Position3D(*data["position"]),
            kind=cube_kind_from_string(data["kind"]),
            label=data["label"],
        )



from __future__ import annotations

from collections.abc import Iterable, Mapping
from itertools import chain

from typing_extensions import override

from tqec.compile.blocks.enums import SpatialBlockBorder, TemporalBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D, round_or_fail


class RepeatedLayer(BaseComposedLayer):
    def __init__(
        self,
        internal_layer: BaseLayer | BaseComposedLayer,
        repetitions: LinearFunction,
        trimmed_spatial_borders: frozenset[SpatialBlockBorder] = frozenset(),
    ):
        """Composed layer implementing repetition.

        This composed layer repeats another layer (that can be atomic or composed)
        multiple times.

        Args:
            internal_layer: repeated layer.
            repetitions: number of repetitions to perform. Can scale with ``k``.
            trimmed_spatial_borders: all the spatial borders that have been
                removed from the layer.

        Raises:
            TQECError: if the total number of timesteps is not a linear
                function (i.e., ``internal_layer.scalable_timesteps.slope != 0``
                and ``repetitions.slope != 0``).
            TQECError: if the total number of timesteps is strictly
                decreasing.

        """
        super().__init__(trimmed_spatial_borders)
        self._internal_layer = internal_layer
        self._repetitions = repetitions
        self._post_init_check()

    @property
    def internal_layer(self) -> BaseLayer | BaseComposedLayer:
        """Get the internal layer that is being repeated by ``self``."""
        return self._internal_layer

    @property
    def repetitions(self) -> LinearFunction:
        """Get the number of repetitions of the internal layer."""
        return self._repetitions

    def _post_init_check(self) -> None:
        # Check that the number of timesteps of self is a linear function.
        if self.internal_layer.scalable_timesteps.slope != 0 and self.repetitions.slope != 0:
            raise TQECError(
                "Layers with a non-constant number of timesteps cannot be "
                "repeated a non-constant number of times as that would lead to "
                "a non-linear number of timesteps, which is not supported yet. "
                f"Got a layer with {self.internal_layer.scalable_timesteps} timesteps "
                f"and tried to repeat it {self.repetitions} times."
            )
        # Check that the number of timesteps of ``self`` is not strictly decreasing.
        if self.repetitions.slope < 0 or self.internal_layer.scalable_timesteps.slope < 0:
            raise TQECError(
                f"Cannot create a {RepeatedLayer.__name__} instance with a decreasing "
                f"number of timesteps. Got repeated layer with "
                f"{self.internal_layer.scalable_timesteps} timesteps that is repeated "
                f"{self.repetitions} times, that would lead to a total duration of "
                f"{self.scalable_timesteps}, which is strictly decreasing."
            )

    @property
    @override
    def scalable_timesteps(self) -> LinearFunction:
        if self.repetitions.slope == 0:
            return self.repetitions.offset * self.internal_layer.scalable_timesteps
        else:
            return self.repetitions * self.internal_layer.scalable_timesteps.offset

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        return self.internal_layer.scalable_shape

    @override
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> RepeatedLayer:
        return RepeatedLayer(
            self.internal_layer.with_spatial_borders_trimmed(borders),
            self.repetitions,
            self.trimmed_spatial_borders | frozenset(borders),
        )

    @staticmethod
    def _get_replaced_layer(
        initial_layer: BaseLayer | BaseComposedLayer,
        border: TemporalBlockBorder,
        border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None],
    ) -> BaseLayer | BaseComposedLayer | None:
        ret: BaseLayer | BaseComposedLayer | None = initial_layer
        if border in border_replacements:
            ret = initial_layer.with_temporal_borders_replaced(
                {border: border_replacements[border]}
            )
        return ret

    @override
    def with_temporal_borders_replaced(
        self, border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None]
    ) -> BaseLayer | BaseComposedLayer | None:
        if not border_replacements:
            return self
        # Compute the bulk layer
        num_borders = len(border_replacements)
        bulk_repetitions = self.repetitions - num_borders
        bulk_layer: BaseLayer | BaseComposedLayer | None
        if bulk_repetitions == LinearFunction(0, 0):
            bulk_layer = None
        elif bulk_repetitions == LinearFunction(0, 1):
            bulk_layer = self.internal_layer
        else:
            bulk_layer = RepeatedLayer(self.internal_layer, bulk_repetitions)

        # Compute the layers at each temporal sides
        initial_layer = self._get_replaced_layer(
            self.internal_layer, TemporalBlockBorder.Z_NEGATIVE, border_replacements
        )
        final_layer = self._get_replaced_layer(
            self.internal_layer, TemporalBlockBorder.Z_POSITIVE, border_replacements
        )
        # Build the resulting layer sequence
        layer_sequence = []
        if (
            # the initial layer has not been removed by the replacement
            initial_layer is not None
            # and it was replaced (i.e., not already accounted for in bulk_layer)
            and TemporalBlockBorder.Z_NEGATIVE in border_replacements
        ):
            layer_sequence.append(initial_layer)
        if bulk_layer is not None:
            layer_sequence.append(bulk_layer)
        if (
            # the final layer has not been removed by the replacement
            final_layer is not None
            # and it was replaced (i.e., not already accounted for in bulk_layer)
            and TemporalBlockBorder.Z_POSITIVE in border_replacements
        ):
            layer_sequence.append(final_layer)
        match len(layer_sequence):
            case 0:
                return None
            case 1:
                return layer_sequence[0]
            case _:
                return SequencedLayers(layer_sequence)

    @override
    def all_layers(self, k: int) -> Iterable[BaseLayer]:
        yield from chain.from_iterable(
            (
                (self.internal_layer,)
                if isinstance(self.internal_layer, BaseLayer)
                else self.internal_layer.all_layers(k)
            )
            for _ in range(self.repetitions.integer_eval(k))
        )

    @override
    def to_sequenced_layer_with_schedule(
        self, schedule: tuple[LinearFunction, ...]
    ) -> SequencedLayers:
        duration = sum(schedule, start=LinearFunction(0, 0))
        if self.scalable_timesteps != duration:
            raise TQECError(
                f"Cannot transform the {RepeatedLayer.__name__} instance to a "
                f"{SequencedLayers.__name__} instance with the provided schedule. "
                f"The provided schedule has a duration of {duration} but the "
                f"instance to transform has a duration of {self.scalable_timesteps}."
            )
        body_duration_scalable = self.internal_layer.scalable_timesteps
        if not body_duration_scalable.is_constant():
            raise NotImplementedError(
                f"Splitting a {RepeatedLayer.__name__} instance with a "
                "non-constant duration body is not implemented yet."
            )
        body_duration = round_or_fail(body_duration_scalable.offset)
        layers: list[BaseLayer | BaseComposedLayer] = []
        for s in schedule:
            try:
                repetitions = s.exact_integer_div(body_duration)
            except TQECError as e:
                raise NotImplementedError(
                    f"The ability to split the body of a {RepeatedLayer.__name__} "
                    "instance has not been implemented yet. Trying to fit an "
                    f"integer repetition of a body with a duration of {body_duration} "
                    f"within a total duration of {s}."
                ) from e
            layers.append(
                RepeatedLayer(self.internal_layer, repetitions)
                if repetitions != LinearFunction(0, 1)
                else self.internal_layer
            )
        return SequencedLayers(layers)

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, RepeatedLayer)
            and self.repetitions == value.repetitions
            and self.internal_layer == value.internal_layer
        )

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    @override
    def get_temporal_layer_on_border(self, border: TemporalBlockBorder) -> BaseLayer:
        return self.internal_layer.get_temporal_layer_on_border(border)

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return LinearFunction.safe_mul(self.internal_layer.scalable_num_moments, self.repetitions)



from __future__ import annotations

import hashlib
import warnings
from collections.abc import Callable, Collection, Iterable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, Literal, cast

import stim

from tqec.circuit.qubit import GridQubit
from tqec.circuit.schedule import ScheduledCircuit
from tqec.plaquette.debug import PlaquetteDebugInformation
from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.qubit import PlaquetteQubits
from tqec.utils.exceptions import TQECError, TQECWarning
from tqec.utils.frozendefaultdict import FrozenDefaultDict
from tqec.utils.position import PhysicalQubitPosition2D


@dataclass(frozen=True)
class Plaquette:
    """Represent a QEC plaquette.

    This class stores qubits in the plaquette local coordinate system and a
    scheduled circuit that should be applied on those qubits to perform the
    QEC experiment.

    By convention, the local plaquette coordinate system is composed of a
    X-axis pointing to the right and a Y-axis pointing down.

    Attributes:
        name: a unique name for the plaquette. This field is used to compare
            plaquettes efficiently (two :class:`Plaquette` instances are
            considered equal if and only if their names are the same) as well as
            computing a hash value for any plaquette instance. Finally, the name
            is used to represent a :class:`Plaquette` instance as a string.
        qubits: qubits used by the plaquette circuit, given in the local
            plaquette coordinate system.
        circuit: scheduled quantum circuit implementing the computation that
            the plaquette should represent.
        mergeable_instructions: a set of instructions that can
            be merged. This is useful when merging several plaquettes'
            circuits together to remove duplicate instructions.

    Raises:
        TQECError: if the provided `circuit` uses qubits not listed in
            `qubits`.

    """

    name: str
    qubits: PlaquetteQubits
    circuit: ScheduledCircuit
    mergeable_instructions: frozenset[str] = field(default_factory=frozenset)
    debug_information: PlaquetteDebugInformation = field(default_factory=PlaquetteDebugInformation)

    def __post_init__(self) -> None:
        plaquette_qubits = set(self.qubits)
        circuit_qubits = set(self.circuit.qubits)
        if not circuit_qubits.issubset(plaquette_qubits):
            wrong_qubits = circuit_qubits.difference(plaquette_qubits)
            raise TQECError(
                f"The following qubits ({wrong_qubits}) are in the provided circuit "
                "but not in the provided list of qubits."
            )

    @property
    def origin(self) -> PhysicalQubitPosition2D:
        """Return the plaquette origin.

        By definition, for the moment, the plaquette origin is always ``(0, 0)``.

        """
        return PhysicalQubitPosition2D(0, 0)

    def __eq__(self, rhs: object) -> bool:
        return isinstance(rhs, Plaquette) and self.name == rhs.name

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return self.name  # pragma: no cover

    def project_on_data_qubit_indices(self, data_qubits_indices: list[int]) -> Plaquette:
        """Project the plaquette on the provided qubit indices.

        This method is useful for deriving a boundary plaquette from a integral
        plaquette.

        Args:
            data_qubits_indices: data-qubit indices that will be kept in the returned plaquette.

        Returns:
            A new plaquette with projected qubits and circuit. The qubits are
            updated to only keep the provided ``data_qubits_indices``. The
            circuit is also updated to only use the kept qubits and empty
            moments with the corresponding schedules are removed.

        """
        kept_qubits = [
            q for i, q in self.qubits.data_qubits_with_indices if i in data_qubits_indices
        ]
        return self.project_on_data_qubits(kept_qubits)

    def project_on_data_qubits(self, data_qubits: list[GridQubit]) -> Plaquette:
        """Project the plaquette on the provided qubits.

        This method is useful for deriving a boundary plaquette from a integral
        plaquette.

        Args:
            data_qubits: qubits that will be kept in the returned plaquette.

        Returns:
            A new plaquette with projected qubits and circuit. The qubits are
            updated to only keep the provided ``data_qubits``. The circuit is
            also updated to only use the kept qubits and empty moments with the
            corresponding schedules are removed.

        """
        removed_data_qubit_indices = [
            i for i, q in self.qubits.data_qubits_with_indices if q not in data_qubits
        ]
        new_plaquette_qubits = PlaquetteQubits(data_qubits, self.qubits.syndrome_qubits)
        new_scheduled_circuit = self.circuit.filter_by_qubits(new_plaquette_qubits.all_qubits)
        debug_info = self.debug_information
        if debug_info is not None:
            debug_info = debug_info.with_data_qubits_removed(removed_data_qubit_indices)

        return Plaquette(
            f"{self.name}_" + "_".join(map(str, removed_data_qubit_indices)),
            new_plaquette_qubits,
            new_scheduled_circuit,
            self.mergeable_instructions,
            debug_info,
        )

    def project_on_boundary(self, projected_orientation: PlaquetteOrientation) -> Plaquette:
        """Project the plaquette on a boundary and return the projected version.

        This method is useful for deriving a boundary plaquette from a integral
        plaquette.

        Args:
            projected_orientation: the orientation of the plaquette after the
                projection.

        Returns:
            A new plaquette with projected qubits and circuit. The qubits are
            updated to only keep the qubits on the side complementary to the
            provided orientation. The circuit is also updated to only use the
            kept qubits and empty moments with the corresponding schedules are
            removed.

        """
        kept_data_qubits = self.qubits.get_qubits_on_side(projected_orientation.to_plaquette_side())
        new_plaquette_qubits = PlaquetteQubits(kept_data_qubits, self.qubits.syndrome_qubits)
        new_scheduled_circuit = self.circuit.filter_by_qubits(new_plaquette_qubits.all_qubits)
        debug_info = self.debug_information.project_on_boundary(projected_orientation)
        return Plaquette(
            f"{self.name}_{projected_orientation.name}",
            new_plaquette_qubits,
            new_scheduled_circuit,
            self.mergeable_instructions,
            debug_info,
        )

    def reliable_hash(self) -> int:
        """Return a hash of ``self`` that is replicable across OSes and Python runs.

        The default ``hash`` function from Python is not guaranteed to always give the same value
        from the same input. In particular, with CPython, ``str`` instances are hashed with an
        additional salt that is randomly picked at each call to the Python executable.

        This is an issue when trying to store a dictionary involving a plaquette as key and re-use
        it in subsequent runs or on other machines / OSes.

        """
        return int(hashlib.md5(self.name.encode()).hexdigest(), 16)  # pragma: no cover

    @property
    def num_measurements(self) -> int:
        """Return the number of measurements performed in the circuit representing ``self``."""
        return self.circuit.num_measurements

    @property
    def num_moments(self) -> int:
        """Return the number of moments in the circuit representing ``self``."""
        return self.circuit.schedule.max_schedule + 1

    def reschedule_measurements(self, schedule: int) -> None:
        """Re-schedule the measurements in the plaquette circuit to the provided ``schedule``.

        Args:
            schedule: new schedule at which measurements should be performed.

        Raises:
            TQECError: if the provided ``schedule`` is earlier than the current
                maximum schedule of the circuit.

        """
        cur_max_schedule = self.circuit.schedule.max_schedule
        if schedule < cur_max_schedule:
            raise TQECError(
                "Cannot reschedule measurements to an earlier time step than the "
                "current maximum schedule."
            )
        # If there is no measurement or if the schedule is the same, do nothing
        if (
            self.circuit.moment_at_schedule(cur_max_schedule).num_measurements == 0
            or schedule == cur_max_schedule
        ):
            return
        self.circuit.reschedule_moment(cur_max_schedule, schedule)

    def is_empty(self) -> bool:
        """Check if the plaquette is empty.

        An empty plaquette is a plaquette that contain empty scheduled circuit.

        """
        return bool(self.circuit.get_circuit(include_qubit_coords=False) == stim.Circuit())

    def with_debug_information(self, debug_information: PlaquetteDebugInformation) -> Plaquette:
        """Create a copy of ``self`` with the provided ``debug_information``."""
        return Plaquette(  # pragma: no cover
            self.name,
            self.qubits,
            self.circuit,
            self.mergeable_instructions,
            debug_information,
        )

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the plaquette.

        The dictionary is intended to be used as a JSON object.

        """
        return {
            "name": self.name,
            "qubits": self.qubits.to_dict(),
            "circuit": self.circuit.to_dict(),
            "mergeable_instructions": list(self.mergeable_instructions),
            "debug_information": self.debug_information.to_dict(),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Plaquette:
        """Return a plaquette from its dictionary representation.

        Args:
            data: dictionary with the keys ``name``, ``qubits``, ``circuit``,
                ``mergeable_instructions`` and ``debug_information``.

        Returns:
            a new instance of :class:`Plaquette` with the provided
            ``name``, ``qubits``, ``circuit``, ``mergeable_instructions`` and
            ``debug_information``.

        """
        name = data["name"]
        qubits = PlaquetteQubits.from_dict(data["qubits"])
        circuit = ScheduledCircuit.from_dict(data["circuit"])
        mergeable_instructions = frozenset(data["mergeable_instructions"])
        debug_information = PlaquetteDebugInformation.from_dict(data["debug_information"])
        return Plaquette(name, qubits, circuit, mergeable_instructions, debug_information)


@dataclass(frozen=True)
class Plaquettes:
    """Represent a collection of plaquettes that might be applied to a :class:`Template` instance.

    The goal of this class is to abstract away how a "collection of plaquettes" is represented and
    to provide a unique interface in order to retrieve plaquettes when building a quantum circuit
    from a template and plaquettes.

    It also checks that the represented collection is valid, which means that it does not include
    any plaquette associated with index 0 (that is internally and conventionally reserved for the
    empty plaquette).

    """

    collection: FrozenDefaultDict[int, Plaquette]

    def __post_init__(self) -> None:
        if 0 in self.collection:
            raise TQECError(
                "Found a Plaquette with index 0. This index is reserved to express "
                '"no plaquette". Please re-number your plaquettes starting from 1.'
            )

    def __getitem__(self, index: int) -> Plaquette:
        return self.collection[index]

    def with_updated_plaquettes(self, plaquettes_to_update: Mapping[int, Plaquette]) -> Plaquettes:
        """Return a new :class:`.Plaquettes` instance with the updated plaquettes.

        Args:
            plaquettes_to_update: a mapping from plaquette indices to :class:`.Plaquette` instances.
                Each entry will replace an existing (or create a new) entry in a copy of ``self``.

        Returns:
            a new :class:`.Plaquettes` instance with plaquettes from ``self`` and
            ``plaquettes_to_update``. If an index is present in both ``self`` and
            ``plaquettes_to_update``, ``plaquettes_to_update`` take precedence.

        """
        return Plaquettes(self.collection | plaquettes_to_update)

    def map_indices(self, callable: Callable[[int], int]) -> Plaquettes:
        """Return a new :class:`.Plaquettes` instance with updated plaquette indices."""
        return Plaquettes(self.collection.map_keys(callable))

    def __eq__(self, rhs: object) -> bool:
        return isinstance(rhs, Plaquettes) and self.collection == rhs.collection

    def __hash__(self) -> int:
        """Implement hashing for Python's hash().

        The returned value is reliable across runs, interpreters and OSes.

        The returned value is reliable across runs, interpreters and OSes.

        """
        return hash(  # pragma: no cover
            tuple(
                sorted(
                    (index, plaquette.reliable_hash())
                    for index, plaquette in self.collection.items()
                )
            )
        )

    def to_name_dict(self) -> dict[int | Literal["default"], str]:
        """Return a dictionary with each plaquette name as value."""
        d: dict[int | Literal["default"], str] = {k: p.name for k, p in self.collection.items()}
        if self.collection.default_value is not None:
            d["default"] = self.collection.default_value.name
        return d

    def without_plaquettes(self, indices: Collection[int]) -> Plaquettes:
        """Remove the plaquettes at the provided ``indices`` from a copy of ``self``."""
        return Plaquettes(
            FrozenDefaultDict(
                {k: v for k, v in self.collection.items() if k not in indices},
                default_value=self.collection.default_value,
            )
        )

    def items(self) -> Iterable[tuple[int, Plaquette]]:
        """Return an iterable over each :class:`.Plaquette` and its associated index in ``self``."""
        return self.collection.items()  # pragma: no cover

    def to_dict(self, plaquettes_to_indices: dict[Plaquette, int] | None = None) -> dict[str, Any]:
        """Return a dictionary representation of the plaquettes.

        Args:
            plaquettes_to_indices: a dictionary mapping plaquettes to their
                indices. If provided, a plaquette will be represented by its index

        """

        def convert(value: Plaquette) -> Any:
            return plaquettes_to_indices[value] if plaquettes_to_indices else value.to_dict()

        return {
            "plaquettes": [
                {"index": index, "plaquette": convert(plaquette)}
                for index, plaquette in self.collection.items()
            ],
            "default": (
                convert(self.collection.default_value)
                if self.collection.default_value is not None
                else None
            ),
        }

    @staticmethod
    def from_dict(
        data: dict[str, Any], plaquettes: Sequence[Plaquette] | None = None
    ) -> Plaquettes:
        """Return a collection of plaquettes from its dictionary representation.

        Args:
            data: dictionary with the keys ``plaquettes`` and ``default``.
            plaquettes: list of :class:`Plaquette` instances to use to build the
                :class:`Plaquettes` instances. Each plaquette is represented by
                its index in the list of unique plaquettes to save space.

        Returns:
            a new instance of :class:`Plaquettes` with the provided
            ``plaquettes`` and ``default``.

        """

        def convert(item: dict[str, Any]) -> Plaquette:
            return (
                Plaquette.from_dict(item["plaquette"])
                if plaquettes is None
                else plaquettes[item["plaquette"]]
            )

        collection = cast(
            FrozenDefaultDict[int, Plaquette],
            FrozenDefaultDict(
                {int(item["index"]): convert(item) for item in data["plaquettes"]},
                default_value=(
                    (Plaquette.from_dict(data["default"]) if data["default"] else None)
                    if plaquettes is None
                    else (plaquettes[data["default"]] if data["default"] is not None else None)
                ),
            ),
        )
        # If the default value is None, print a WARNING (this should not happen in practice)
        if collection.default_value is None:
            warnings.warn(
                "The default value of the plaquettes collection is None. "
                "This should not happen in practice.",
                TQECWarning,
            )
        return Plaquettes(collection)



"""Defines :func:`~.compile.compile_block_graph`."""

from typing import Final, Literal

from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.repeated import RepeatedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.convention import FIXED_BULK_CONVENTION, Convention
from tqec.compile.graph import TopologicalComputationGraph
from tqec.compile.observables.abstract_observable import (
    AbstractObservable,
    compile_correlation_surface_to_abstract_observable,
)
from tqec.compile.specs.base import CubeSpec, PipeSpec
from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface
from tqec.computation.cube import Cube
from tqec.templates.base import RectangularTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import BlockPosition3D, Direction3D
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D

_DEFAULT_SCALABLE_QUBIT_SHAPE: Final = PhysicalQubitScalable2D(
    LinearFunction(4, 5), LinearFunction(4, 5)
)

_DEFAULT_BLOCK_REPETITIONS: LinearFunction = LinearFunction(2, -1)


def _get_template_from_layer(
    root: BaseLayer | BaseComposedLayer,
) -> RectangularTemplate:
    """Get a unique template from any given layer.

    This helper function try its best to recover the template a given layer uses.

    Raises:
        TQECError: if an instance of :class:`.BaseLayer` is something else than an instance of
            :class:`.PlaquetteLayer`, because :class:`.PlaquetteLayer` is the only class from which
            we can recover a template instance.
        TQECError: if an instance of :class:`.SequencedLayers` contains sub-layers with
            different templates.
        NotImplementedError: if an unknown layer is found.

    Returns:
        the template used by the provided ``root`` layer.

    """
    if isinstance(root, BaseLayer):
        if not isinstance(root, PlaquetteLayer):
            raise TQECError(
                f"Trying to get the Template from a {type(root).__name__} "
                "instance that does not have any Template."
            )
        return root.template
    elif isinstance(root, SequencedLayers):
        possible_templates = {_get_template_from_layer(layer) for layer in root.layer_sequence}
        if len(possible_templates) > 1:
            raise TQECError(
                "Multiple possible Template found:\n  -"
                + "\n  -".join(type(t).__name__ for t in possible_templates)
                + "\nWhich is not supported at the moment."
            )
        return next(iter(possible_templates))
    elif isinstance(root, RepeatedLayer):
        return _get_template_from_layer(root.internal_layer)
    else:
        raise NotImplementedError("Unknown layer type encountered:", type(root).__name__)


def compile_block_graph(
    block_graph: BlockGraph,
    convention: Convention = FIXED_BULK_CONVENTION,
    observables: list[CorrelationSurface] | Literal["auto"] | None = "auto",
    block_temporal_height: LinearFunction = _DEFAULT_BLOCK_REPETITIONS,
) -> TopologicalComputationGraph:
    """Compile a block graph.

    Args:
        block_graph: The block graph to compile.
        convention: convention used to generate the quantum circuits.
        observables: correlation surfaces that should be compiled into
            observables and included in the compiled circuit.
            If set to ``"auto"``, the correlation surfaces will be automatically
            determined from the block graph. If a list of correlation surfaces
            is provided, only those surfaces will be compiled into observables
            and included in the compiled circuit. If set to ``None``, no
            observables will be included in the compiled circuit.
        block_temporal_height: the number of rounds of stabilizer measurements
            (ignoring one layer for initialization and another for final measurement).
            Defaults to `2k-1`.

    Returns:
        A :class:`TopologicalComputationGraph` object that can be used to generate a
        ``stim.Circuit`` and scale easily.

    """
    # All the ports should be filled before compiling the block graph.
    if block_graph.num_ports != 0:
        raise TQECError(
            "Can not compile a block graph with open ports into circuits. "
            "You might want to call `fill_ports` or `fill_ports_for_minimal_simulation` "
            "on the block graph before compiling it."
        )
    # Validate the graph can represent a valid computation.
    block_graph.validate()

    # Fix the shadowed faces of the cubes to avoid using spatial cubes
    # when a non-spatial cube can be used at the same position.
    # For example, when three XXZ cubes are connected in a row along the x-axis,
    # the middle one can be replaced by a ZXX cube because the faces along the
    # x-axis are shadowed by the connected pipes.
    block_graph = block_graph.fix_shadowed_faces()

    # Set the minimum z of block graph to 0.(time starts from zero)
    minz = min(cube.position.z for cube in block_graph.cubes)
    if minz != 0:
        block_graph = block_graph.shift_by(dz=-minz)

    # We need to know exactly which spatial pipes will be placed on a time slice where extended
    # plaquettes will be used, in order to adapt the schedule of the measurement layer.
    def has_pipes_in_both_spatial_dimensions(cube: Cube) -> bool:
        return frozenset(
            pipe.direction for pipe in block_graph.pipes_at(cube.position) if pipe.kind.is_spatial
        ) == frozenset([Direction3D.X, Direction3D.Y])

    extended_stabilizers_pipe_slices: frozenset[int] = frozenset(
        pipe.u.position.z
        for pipe in block_graph.pipes
        if (
            pipe.direction == Direction3D.Y
            and (
                has_pipes_in_both_spatial_dimensions(pipe.u)
                ^ has_pipes_in_both_spatial_dimensions(pipe.v)
            )
        )
    )
    cube_specs = {
        cube: CubeSpec.from_cube(cube, block_graph, extended_stabilizers_pipe_slices)
        for cube in block_graph.cubes
    }

    # 0. Get the abstract observables to be included in the compiled circuit.
    obs_included: list[AbstractObservable] = []
    if observables is not None:
        if observables == "auto":
            observables = block_graph.find_correlation_surfaces()
        include_temporal_hadamard_pipes = convention.name == "fixed_bulk"
        obs_included = [
            compile_correlation_surface_to_abstract_observable(
                block_graph, surface, include_temporal_hadamard_pipes
            )
            for surface in observables
        ]

    # 1. Create topological computation graph
    graph = TopologicalComputationGraph(
        _DEFAULT_SCALABLE_QUBIT_SHAPE,
        observables=obs_included,
        observable_builder=convention.triplet.observable_builder,
    )

    # 2. Add cubes to the graph
    for cube in block_graph.cubes:
        spec = cube_specs[cube]
        position = BlockPosition3D(cube.position.x, cube.position.y, cube.position.z)
        graph.add_cube(position, convention.triplet.cube_builder(spec, block_temporal_height))

    # 3. Add pipes to the graph
    # Note that the order of the pipes to add is important.
    # To keep the time-direction pipes from removing the extra resets
    # added by the space-direction pipes, we first add the time-direction pipes
    pipes = block_graph.pipes
    time_pipes = [pipe for pipe in pipes if pipe.direction == Direction3D.Z]
    temporal_hadamard_z_positions: set[int] = {
        pipe.u.position.z for pipe in time_pipes if pipe.kind.has_hadamard
    }
    space_pipes = [pipe for pipe in pipes if pipe.direction != Direction3D.Z]
    for pipe in time_pipes + space_pipes:
        pos1, pos2 = pipe.u.position, pipe.v.position
        pos1 = BlockPosition3D(pos1.x, pos1.y, pos1.z)
        pos2 = BlockPosition3D(pos2.x, pos2.y, pos2.z)
        template1 = _get_template_from_layer(graph.get_cube(pos1))
        template2 = _get_template_from_layer(graph.get_cube(pos2))
        key = PipeSpec(
            (cube_specs[pipe.u], cube_specs[pipe.v]),
            (template1, template2),
            pipe.kind,
            has_spatial_up_or_down_pipe_in_timeslice=(
                pos1.z == pos2.z and pos1.z in extended_stabilizers_pipe_slices
            ),
            at_temporal_hadamard_layer=(
                pipe.kind.is_temporal and pos1.z in temporal_hadamard_z_positions
            ),
        )
        graph.add_pipe(pos1, pos2, convention.triplet.pipe_builder(key, block_temporal_height))

    return graph



from __future__ import annotations

import argparse
import os
from pathlib import Path

import matplotlib.pyplot as plt
from typing_extensions import override

from tqec._cli.subcommands.base import TQECSubCommand
from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import CorrelationSurface
from tqec.interop.pyzx.plot import (
    draw_correlation_surface_on,
    draw_positioned_zx_graph_on,
)
from tqec.interop.pyzx.positioned import PositionedZX


class Dae2ObservablesTQECSubCommand(TQECSubCommand):
    @staticmethod
    @override
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser: argparse.ArgumentParser = main_parser.add_parser(
            "dae2observables",
            description="Takes a .dae file in and list all the observables that has been found.",
        )
        parser.add_argument(
            "dae_file", help="A valid .dae file representing a computation.", type=Path
        )
        parser.add_argument(
            "--out-dir",
            help="An optional argument providing the directory in which to "
            "export images representing the observables found.",
            type=Path,
        )
        parser.set_defaults(func=Dae2ObservablesTQECSubCommand.execute)

    @staticmethod
    @override
    def execute(args: argparse.Namespace) -> None:
        dae_absolute_path: Path = args.dae_file.resolve()
        block_graph = BlockGraph.from_dae_file(dae_absolute_path, graph_name=str(dae_absolute_path))
        zx_graph = block_graph.to_zx_graph()
        correlation_surfaces = block_graph.find_correlation_surfaces()

        if args.out_dir is None:
            print(
                f"Found {len(correlation_surfaces)} observables. Please provide an output "
                "directory by using the '--out-dir' argument to visualize them."
            )
        else:
            if not args.out_dir.exists():
                os.makedirs(args.out_dir)
            save_correlation_surfaces_to(zx_graph, args.out_dir, correlation_surfaces)


def save_correlation_surfaces_to(
    zx_graph: PositionedZX,
    out_dir: Path,
    correlation_surfaces: list[CorrelationSurface],
) -> None:
    """Save the provided correlation surfaces to ``out_dir``.

    Args:
        zx_graph: ZX graph supporting the provided ``correlation_surfaces``. Plotted as background
            so that correlation surfaces appear on the ZX-graph.
        out_dir: filepath to save the drawing to.
        correlation_surfaces: correlation surfaces to draw over ``zx_graph``.

    """
    for i, correlation_surface in enumerate(correlation_surfaces):
        fig = plt.figure(figsize=(5, 6))
        ax = fig.add_subplot(111, projection="3d")
        draw_positioned_zx_graph_on(zx_graph, ax)
        draw_correlation_surface_on(correlation_surface, zx_graph, ax)
        fig.tight_layout()
        save_path = (out_dir / f"{i}.png").resolve()
        print(f"Saving correlation surface number {i} to '{save_path}'.")
        fig.savefig(save_path)
        fig.clear()



"""Defines the base classes for templates: :class:`Template` and :class:`RectangularTemplate`."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from typing import Any

import numpy
import numpy.typing as npt

from tqec.templates.enums import TemplateBorder
from tqec.templates.subtemplates import (
    UniqueSubTemplates,
    get_spatially_distinct_subtemplates,
)
from tqec.utils.position import (
    BlockPosition2D,
    PlaquettePosition2D,
    PlaquetteShape2D,
    Shift2D,
)
from tqec.utils.scale import PlaquetteScalable2D, round_or_fail


class Template(ABC):
    """Base class for all the templates.

    This class is the base of all templates and provide the necessary interface that all templates
    should implement to be usable by the library.

    """

    def __init__(self, default_increments: Shift2D | None = None) -> None:
        """Construct an instance of the template.

        Args:
            default_increments: default increments between two plaquettes. Defaults
                to ``Displacement(2, 2)`` when ``None``

        """
        super().__init__()
        self._default_shift = default_increments or Shift2D(2, 2)

    def __hash__(self) -> int:
        return hash((type(self), self._default_shift))

    def __eq__(self, value: Any) -> bool:
        return type(self) is type(value) and self._default_shift == value._default_shift

    @abstractmethod
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        """Generate the numpy array representing the template.

        Args:
            k: scaling parameter used to instantiate the template.
            plaquette_indices: the plaquette indices that will be forwarded to
                the underlying Shape instance's instantiate method. Defaults
                to ``range(1, self.expected_plaquettes_number + 1)`` if ``None``.

        Returns:
            a numpy array with the given plaquette indices arranged according to
            the underlying shape of the template.

        """

    def instantiate_list(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> list[list[int]]:
        """Generate a 2-dimensional list of integers representing the template.

        This method is equivalent to
        ``self.instantiate(k, plaquette_indices).tolist()`` but has a stricter
        and more correct typing than calling ``tolist`` on a numpy array.

        Args:
            k: scaling parameter used to instantiate the template.
            plaquette_indices: the plaquette indices that will be forwarded to
                the underlying Shape instance's instantiate method. Defaults
                to ``range(1, self.expected_plaquettes_number + 1)`` if ``None``.

        Returns:
            a 2-dimensional list (i.e., a list of lists) with the given
            plaquette indices arranged according to the underlying shape of the
            template.

        """
        instantiation = self.instantiate(k, plaquette_indices=plaquette_indices)
        m, n = instantiation.shape
        ret: list[list[int]] = []
        for i in range(m):
            ret.append([])
            for j in range(n):
                ret[-1].append(int(instantiation[i, j]))
        return ret

    def shape(self, k: int) -> PlaquetteShape2D:
        """Return the current template shape."""
        sshape = self.scalable_shape
        return PlaquetteShape2D(round_or_fail(sshape.x(k)), round_or_fail(sshape.y(k)))

    @property
    @abstractmethod
    def scalable_shape(self) -> PlaquetteScalable2D:
        """Return a scalable version of the template shape."""

    @property
    @abstractmethod
    def expected_plaquettes_number(self) -> int:
        """Return the number of plaquettes expected from the :py:meth:`instantiate` method.

        Returns:
            the number of plaquettes expected from the :py:meth:`instantiate` method.

        """

    def get_increments(self) -> Shift2D:
        """Get the default increments of the template.

        Returns:
            a displacement of the default increments in the x and y directions.

        """
        return self._default_shift

    def get_spatially_distinct_subtemplates(
        self, k: int, manhattan_radius: int = 1, avoid_zero_plaquettes: bool = True
    ) -> UniqueSubTemplates:
        """Return a representation of the distinct sub-templates of the provided Manhattan radius.

        Note:
            This method will likely be inefficient for large templates (i.e., large
            values of `k`) or for large Manhattan radiuses, both in terms of memory
            used and computation time.
            Subclasses are invited to reimplement that method using a specialized
            algorithm (or hard-coded values) to speed things up.

        Args:
            k: scaling parameter used to instantiate the template.
            manhattan_radius: radius of the considered ball using the Manhattan
                distance. Only squares with sides of ``2*manhattan_radius+1``
                plaquettes will be considered.
            avoid_zero_plaquettes: ``True`` if sub-templates with an empty plaquette
                (i.e., 0 value in the instantiation of the
                :class:`~tqec.templates.base.Template` instance) at its center
                should be ignored. Default to ``True``.

        Returns:
            a representation of all the sub-templates found.

        """
        return get_spatially_distinct_subtemplates(  # pragma: no cover
            self.instantiate(k), manhattan_radius, avoid_zero_plaquettes
        )

    def instantiation_origin(self, k: int) -> PlaquettePosition2D:
        """Coordinates of the top-left entry origin.

        This property returns the coordinates of the origin of the plaquette
        (:class:`~tqec.plaquette.plaquette.Plaquette.origin`) that corresponds
        to the top-left entry of the array returned by
        :meth:`~tqec.templates.base.Template.instantiate`.

        Note:
            the returned coordinates are in plaquette coordinates. That means
            that, if you want to get the coordinates of the top-left plaquette
            origin (which is a qubit), you should multiply the coordinates
            returned by this method by the tiling increments.

        Args:
            k: scaling parameter used to instantiate the template.

        Returns:
            the coordinates of the origin of the plaquette
            (:class:`~tqec.plaquette.plaquette.Plaquette.origin`) that corresponds
            to the top-left entry of the array returned by
            :meth:`~tqec.templates.base.Template.instantiate`.

        """
        return BlockPosition2D(0, 0).get_top_left_plaquette_position(self.shape(k))


@dataclass(frozen=True)
class BorderIndices:
    """Stores indices on the border of a rectangular template.

    Attributes:
        top_left_corner: non-repeating index at the top or left part of the
            border.
        first_repeating: first repeating index on the border "bulk".
        second_repeating: second repeating index on the border "bulk".
        bottom_right_corner: non-repeating index at the bottom or right part of
            the border.

    """

    top_left_corner: int
    first_repeating: int
    second_repeating: int
    bottom_right_corner: int

    def to(self, other: BorderIndices) -> dict[int, int]:
        """Return a mapping from ``self`` to ``other``.

        This method returns a mapping from the indices stored in ``self`` to the
        indices stored in ``other``.

        Args:
            other: indices from another border.

        Returns:
            a mapping from the indices stored in ``self`` to the indices stored
            in ``other``.

        """
        return {s: o for s, o in zip(self, other)}

    def __iter__(self) -> Iterator[int]:
        yield from (
            self.top_left_corner,
            self.first_repeating,
            self.second_repeating,
            self.bottom_right_corner,
        )


class RectangularTemplate(Template):
    """Base class for all the templates that have a rectangular shape."""

    @abstractmethod
    def get_border_indices(self, border: TemplateBorder) -> BorderIndices:
        """Return the indices on the provided ``border``.

        Args:
            border: side of the template instance for which the indices are
                needed.

        Returns:
            a description of the indices present on the provided ``border`` of
            the represented template.

        """
        pass



from collections.abc import Iterable
from typing import cast

import svg
from typing_extensions import override

from tqec.plaquette.enums import PlaquetteOrientation
from tqec.plaquette.rpng.rpng import RPNG, ExtendedBasis, PauliBasis, RPNGDescription
from tqec.visualisation.computation.plaquette.base import PlaquetteCorner, SVGPlaquetteDrawer, lerp
from tqec.visualisation.configuration import DrawerConfiguration
from tqec.visualisation.exception import TQECDrawingError


def _get_bounding_box(coords: Iterable[complex]) -> tuple[complex, complex]:
    min_x = min(c.real for c in coords)
    max_x = max(c.real for c in coords)
    min_y = min(c.imag for c in coords)
    max_y = max(c.imag for c in coords)
    return complex(min_x, min_y), complex(max_x, max_y)


class RPNGPlaquetteDrawer(SVGPlaquetteDrawer):
    def __init__(self, description: RPNGDescription) -> None:
        """SVG plaquette drawer for RPNG descriptions.

        Args:
            description: :class:`RPNGDescription` that should be drawn by this drawer.

        """
        super().__init__()
        self._description = description
        self._center = RPNGPlaquetteDrawer._CENTER_COORDINATE
        bases = {rpng.p for rpng in description.corners if rpng.p is not None}
        uniform_basis: PauliBasis | None = next(iter(bases)) if len(bases) == 1 else None
        self._uniform_colour = (
            SVGPlaquetteDrawer.get_colour(uniform_basis) if uniform_basis is not None else None
        )
        self._sorted_corner_indices: list[int] = []
        self._corners: list[complex] = []
        self._corners_enum: list[PlaquetteCorner] = []
        self._rpngs: list[RPNG] = []
        for i, (coords, corner_enum, rpng) in enumerate(
            zip(
                RPNGPlaquetteDrawer._CORNERS,
                RPNGPlaquetteDrawer._CORNERS_ENUM,
                description.corners,
            )
        ):
            if rpng.is_null:
                continue
            self._sorted_corner_indices.append(i)
            self._corners.append(coords)
            self._corners_enum.append(corner_enum)
            self._rpngs.append(rpng)
        self._uuid: str = "_".join(
            [
                str(self._description.ancilla),
                *[str(corner) for corner in description.corners],
            ]
        )

    @override
    def draw(
        self,
        id: str,
        show_interaction_order: bool = True,
        show_hook_errors: bool = True,
        show_data_qubit_reset_measurements: bool = True,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        # Return an empty SVG element if there are no corners.
        if len(self._corners) == 0:
            return svg.G(id=id, elements=[])
        # Build iteratively the different layers that we need to represent the
        # plaquette.
        shape_path = self.get_plaquette_shape_path(configuration)
        layers: list[svg.Element] = []
        # If the plaquette has a uniform color, the shape include that color with
        # its "fill" attribute and so we do not need to clip. If that is not the
        # case, the filling colours will be added with get_fill_layer, but it is
        # simpler for the implementation to clip.
        if self._uniform_colour is None:
            layers.append(svg.ClipPath(id=self._uuid, elements=[shape_path]))
            layers.extend(self.get_fill_layers(configuration))
        # Second layer: the strokes (with filling if self._uniform_colour is set)
        layers.append(shape_path)
        if show_data_qubit_reset_measurements:
            layers.append(self.get_data_qubit_reset_measurements_layers(configuration))
        if show_hook_errors:
            if (line := self.get_hook_error_line(configuration)) is not None:
                layers.append(line)
        # Third layer: the text.
        if show_interaction_order:
            layers.extend(self.get_interaction_order_text(configuration))
        return svg.G(id=id, elements=layers)

    def get_plaquette_shape_path(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Element:
        """Return the plaquette shape, filled iff it measures in a uniform Pauli basis.

        This method returns the plaquette shape as an SVG path. It might also fill this shape if
        ``self`` represents a plaquette measuring its qubits in the same Pauli basis, else without
        fill at all.

        Args:
            configuration: drawing configuration.

        Returns:
            The plaquette shape. It is filled with the appropriate colour if the
            stabilizer measured in ``self`` measures in the same Pauli basis for
            all the qubits (e.g. ``XXXX``). If different bases are present (e.g.
            ``ZXXZ``), the returned SVG element does not contain any fill, and
            the colours should be added later.

        """
        fill = "none" if self._uniform_colour is None else self._uniform_colour
        match len(self._corners):
            case 2:
                # Find the plaquette orientation
                orientation: PlaquetteOrientation
                match self._sorted_corner_indices:
                    case [0, 1]:
                        orientation = PlaquetteOrientation.DOWN
                    case [0, 2]:
                        orientation = PlaquetteOrientation.RIGHT
                    case [1, 3]:
                        orientation = PlaquetteOrientation.LEFT
                    case [2, 3]:
                        orientation = PlaquetteOrientation.UP
                    case _:
                        raise TQECDrawingError(
                            "Could not find the orientation of a plaquette containing the "
                            f"corner indices {self._sorted_corner_indices}."
                        )
                # Return the corresponding shape
                return self.get_half_circle_shape(orientation, fill, configuration)

            case 3:
                # Find the plaquette placement
                place: PlaquetteCorner
                match self._sorted_corner_indices:
                    case [1, 2, 3]:
                        place = PlaquetteCorner.TOP_LEFT
                    case [0, 2, 3]:
                        place = PlaquetteCorner.TOP_RIGHT
                    case [0, 1, 3]:
                        place = PlaquetteCorner.BOTTOM_LEFT
                    case [0, 1, 2]:
                        place = PlaquetteCorner.BOTTOM_RIGHT
                    case _:
                        raise TQECDrawingError(
                            "Could not find the placement of a plaquette containing the "
                            f"corner indices {self._sorted_corner_indices}."
                        )
                return self.get_triangle_shape(place, fill, configuration)
            case 4:
                return self.get_square_shape(fill, configuration)
            case num:
                raise TQECDrawingError(
                    f"Got a plaquette with {num} corners. Only 2, 3 or 4 corners are supported."
                )

    def get_fill_layers(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> list[svg.Element]:
        """Return one SVG element per non-empty corners filling each corresponding quarter.

        This method should only be called if the plaquette drawn by self measures data-qubits in a
        non-uniform Pauli basis (e.g. ZXXZ).

        Note:
            The returned filled rectangles are clipped with the clipPath with
            id ``self._uuid`` that should be created before following the
            plaquette shape path.

        Args:
            configuration: drawing configuration.

        Returns:
            A list of ``len(self._corners)`` SVG elements, each filling a
            quarter of the square plaquette and clipping to the clipPath with
            id ``self._uuid``.

        """
        # Draw one rectangle for each corner
        fill_layer: list[svg.Element] = []
        for corner, rpng in zip(self._corners, self._rpngs):
            tl, br = _get_bounding_box([self._center, corner])
            basis = rpng.p
            fill = configuration.mixed_basis_color if basis is None else self.get_colour(basis)
            fill_layer.append(
                svg.Rect(
                    x=tl.real,
                    y=tl.imag,
                    width=(br.real - tl.real),
                    height=(br.imag - tl.imag),
                    fill=fill,
                    stroke=None,
                    clip_path=f"url(#{self._uuid})",
                )
            )
        return fill_layer

    def get_interaction_order_text(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> list[svg.Text]:
        """Return a SVG element containing data-qubit interaction orders as text.

        This function returns one SVG element per non-empty corners, each containing a text element
        with the time slice at which a 2-qubit operation is applied on the corner qubit.

        Args:
            configuration: drawing configuration.

        Returns:
            One SVG element per non-empty corners, each containing a text
            element with the time slice at which a 2-qubit operation is applied
            on the corner qubit.

        """
        interaction_order_texts: list[svg.Text] = []
        for corner, rpng in zip(self._corners, self._rpngs):
            if rpng.n is None:
                continue
            text_position = lerp(self._center, corner, configuration.text_lerp_coefficient)
            interaction_order_texts.append(
                svg.Text(
                    x=text_position.real,
                    y=text_position.imag,
                    fill="black",
                    font_size=0.1,
                    text_anchor="middle",
                    dominant_baseline="central",
                    text=str(rpng.n),
                )
            )
        return interaction_order_texts

    def get_hook_error_line(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.Line | None:
        """Return a SVG line showing the direction of the hook error.

        Args:
            configuration: drawing configuration.

        Returns:
            A SVG line showing the direction of the hook error, or ``None`` if
            there is no hook error.

        """
        if len(self._corners) != 4:
            return None
        sorted_rpngs = sorted([(rpng.n, i) for i, rpng in enumerate(self._rpngs)])
        f = configuration.hook_error_line_lerp_coefficient
        a = lerp(self._center, self._corners[sorted_rpngs[-1][1]], f)
        b = lerp(self._center, self._corners[sorted_rpngs[-2][1]], f)
        return svg.Line(
            x1=a.real,
            x2=b.real,
            y1=a.imag,
            y2=b.imag,
            stroke=configuration.stroke_color,
            stroke_width=configuration.stroke_width,
        )

    def get_data_qubit_reset_measurements_layers(
        self,
        configuration: DrawerConfiguration = DrawerConfiguration(),
    ) -> svg.G:
        """Return a SVG layer containing a representation of data-qubit resets/measurements.

        Args:
            configuration: drawing configuration.

        Returns:
            a SVG element containing all the reset or measurements on the
            plaquette drawn by ``self``.

        """
        reset_measurement_elements: list[svg.Element] = []
        for corner, place, rpng in zip(self._corners, self._corners_enum, self._rpngs):
            r, m = rpng.r, rpng.g
            r_is_basis = r is not None
            m_is_basis = m is not None and m != ExtendedBasis.H
            if not r_is_basis and not m_is_basis:
                continue
            if r_is_basis and m_is_basis:
                raise TQECDrawingError(
                    "Found an RPNG with both reset and measurement. That is not "
                    "supported in the drawer."
                )
            assert r_is_basis ^ m_is_basis
            fill = self.get_colour(r if r is not None else cast(ExtendedBasis, m))
            shape = (
                self.get_reset_shape(place, fill, configuration)
                if r_is_basis
                else self.get_measurement_shape(place, fill, configuration)
            )
            reset_measurement_elements.append(
                svg.G(
                    elements=[shape],
                    transform=[svg.Translate(corner.real, corner.imag)],
                )
            )
        return svg.G(elements=reset_measurement_elements)



"""Block graph that represents a logical stability experiment."""

from tqec.computation.block_graph import BlockGraph, ZXCube
from tqec.utils.enums import Basis
from tqec.utils.paths import GALLERY_DAE_DIR

STEANE_CODE_DAE = GALLERY_DAE_DIR / "steane_encoding.dae"


def steane_encoding(observable_basis: Basis | None = None) -> BlockGraph:
    """Create a block graph that represents a Steane encoding circuit.

    The block graph is created from a DAE file that describes the circuit.

    Args:
        observable_basis: The observable basis that the block graph can support. If
            None, the block graph will have open ports. Otherwise, the ports will be
            filled with the given observable basis.

    Returns:
        A :py:class:`~tqec.computation.block_graph.BlockGraph` instance.

    """
    graph = BlockGraph.from_dae_file(STEANE_CODE_DAE)
    if observable_basis is not None:
        graph.fill_ports(
            {f"Port{port}": ZXCube(Basis.Z, Basis.X, observable_basis) for port in (3, 4)}
        )
        graph.fill_ports(
            {f"Port{port}": ZXCube(Basis.X, Basis.Z, observable_basis) for port in (1, 2, 5, 6)}
        )
        graph.fill_ports(ZXCube(observable_basis, Basis.X, Basis.Z))
    return graph



"""Utilities for working with graphs with open ports."""

import warnings
from collections.abc import Iterator
from dataclasses import dataclass
from functools import reduce
from itertools import combinations
from typing import Final

import networkx as nx
from pyzx.pauliweb import multiply_paulis

from tqec.computation.block_graph import BlockGraph
from tqec.computation.correlation import (
    CorrelationSurface,
    reduce_observables_to_minimal_generators,
)
from tqec.computation.cube import YHalfCube, ZXCube
from tqec.utils.enums import Basis
from tqec.utils.exceptions import TQECError, TQECWarning
from tqec.utils.position import Direction3D


@dataclass(frozen=True)
class FilledGraph:
    """A block graph without open ports that can be used for simulation.

    Attributes:
        graph: The block graph with all ports filled.
        stabilizers: The external stabilizers simulable on the filled graph.
        observables: The correlation surfaces that can be used as logical observables
            for the simulation. A list of independent correlation surfaces with the
            area as small as possible is chosen.

    """

    graph: BlockGraph
    stabilizers: list[str]
    observables: list[CorrelationSurface]

    def __post_init__(self) -> None:
        if self.graph.num_ports != 0:
            raise TQECError("The filled graph should not have open ports.")
        if len(self.stabilizers) != len(self.observables):
            raise TQECError("The number of stabilizers and observables should match.")

    def get_external_stabilizers(self) -> list[str]:
        """Return the external stabilizers of the correlation surfaces."""
        return [obs.external_stabilizer_on_graph(self.graph) for obs in self.observables]


def fill_ports_for_minimal_simulation(
    graph: BlockGraph,
    search_small_area_observables: bool = False,
) -> list[FilledGraph]:
    """Fill the ports of the provided ``graph`` to minimize the number of simulation runs.

    Given a block graph with open ports, fill in the ports with the appropriate cubes that will
    minimize the number of simulation runs needed for the complete logical observable set.

    The key idea is that stabilizers within the same clique should have compatible
    Paulis at their ports. This allows a single simulation run to simultaneously
    simulate for all stabilizers in a clique. Minimizing the number of simulations,
    therefore, translates to minimizing the number of cliques while ensuring full
    coverage of all stabilizers. This leads to solving the minimum clique cover
    problem on a graph, which in turn reduces to a graph coloring problem on its
    complement.

    Args:
        graph: The block graph with open ports.
        search_small_area_observables: If True, the algorithm will try to construct
            all the possible correlation surfaces that can be used as logical
            observables and select the generators with the smallest area. This will
            result in constructing exponentially many correlation surfaces, which
            can be slow for graphs with large number of ports.

    Returns:
        A list of :class:`~tqec.computation.open_graph.FilledGraph` instances, each
        containing a block graph with all ports filled and a set of correlation
        surfaces that can be used as logical observables for the simulation on that
        block graph.

    """
    num_ports = graph.num_ports
    if num_ports == 0:
        raise TQECError("The provided graph has no open ports.")
    # heuristic threshold for large number of ports
    _heuristic_threshold: Final[int] = 16
    if search_small_area_observables and num_ports > _heuristic_threshold:
        warnings.warn(
            "The algorithm will construct all exponentially many correlation "
            "surfaces, which can be slow for graphs with large number of ports. "
            "Consider setting `search_small_area_observables=False` for better "
            "performance.",
            TQECWarning,
        )

    correlation_surfaces = graph.find_correlation_surfaces()
    stab_to_surface: dict[str, CorrelationSurface] = {
        s.external_stabilizer_on_graph(graph): s for s in correlation_surfaces
    }
    generators = list(stab_to_surface.keys())

    if search_small_area_observables:
        identity = "I" * num_ports
        # Need to collect all the possible correlation surfaces because we want
        # to find the generators with the smallest correlation surface area
        for stabilizer, comb in _iter_stabilizer_group(generators):
            if stabilizer != identity and stabilizer not in stab_to_surface:
                correlation_surface = reduce(
                    lambda a, b: a ^ b,
                    [stab_to_surface[s] for s in comb],
                )
                stab_to_surface[stabilizer] = correlation_surface
        generators = list(
            reduce_observables_to_minimal_generators(stab_to_surface, num_ports).keys()
        )

    # Two stabilizers are compatible if they can agree on the supported observable
    # basis on the common ports. We can construct a graph that assigns a node to
    # each stabilizer and an edge between two nodes if the stabilizers are compatible.
    # Then we can solve the minimum clique cover problem to find the minimum
    # number of cliques that cover all the nodes in the graph. Each clique will
    # correspond to a port configuration that can be used for simulation.
    # Here we solve the graph coloring problem on the complement graph, which is
    # equivalent to the clique cover problem on the original graph.
    nodes = list(range(len(generators)))
    g: nx.Graph[int] = nx.Graph()
    g.add_nodes_from(nodes)
    for i, j in combinations(nodes, 2):
        if not _is_compatible_paulis(generators[i], generators[j]):
            g.add_edge(i, j)
    # Solve with heuristic greedy coloring
    coloring = nx.algorithms.coloring.greedy_color(g)  # type: ignore[invalid-argument-type]
    cliques: dict[int, list[str]] = {}
    for node, color in coloring.items():
        cliques.setdefault(color, []).append(generators[node])

    def ports_basis_for_clique(
        supported_stabilizers: list[str],
    ) -> list[str]:
        port_ops = ["I"] * num_ports
        for s in supported_stabilizers:
            for i, p in enumerate(s):
                if p != "I":
                    port_ops[i] = p
        # For the ports with operator I, we can choose any basis
        # Here we choose the Z basis for consistency
        return ["Z" if p == "I" else p for p in port_ops]

    # Fill in the ports and create the filled graphs
    ports = graph.ordered_ports
    filled_graphs: list[FilledGraph] = []
    for clique in cliques.values():
        fg = graph.clone()
        port_basis = ports_basis_for_clique(clique)
        for port, basis in zip(ports, port_basis):
            port_pos = graph.ports[port]
            if basis == "Y":  # pragma: no cover
                fg.fill_ports({port: YHalfCube()})
                continue
            assert basis in ["Z", "X"]
            # Match the basis of the pipe at the port
            pipe_at_port = graph.pipes_at(port_pos)[0]
            pipe_kind = pipe_at_port.kind
            at_head = pipe_at_port.u == graph[port_pos]
            fill_kind = ZXCube(
                *[
                    pipe_kind.get_basis_along(dir, at_head) or Basis(basis)
                    for dir in Direction3D.all_directions()
                ]
            )
            fg.fill_ports({port: fill_kind})
        assert fg.num_ports == 0

        filled_graphs.append(
            FilledGraph(
                graph=fg,
                stabilizers=clique,
                observables=[stab_to_surface[s] for s in clique],
            )
        )
    return filled_graphs


def _multiply_unsigned_paulis(p1: str, p2: str) -> str:
    return "".join(
        multiply_paulis(
            p1[i],
            p2[i],
        )
        for i in range(len(p1))
    )


def _iter_stabilizer_group(
    generators: list[str],
) -> Iterator[tuple[str, tuple[str, ...]]]:
    """Iterate over the stabilizer group generated by the given generators.

    Return the stabilizer and the generators that generate it.

    """
    identity = "I" * len(generators[0])
    # powerset of the generators
    for r in range(1, len(generators) + 1):
        for comb in combinations(generators, r):
            output = identity
            for gen in comb:
                output = _multiply_unsigned_paulis(output, gen)

            yield (output, comb)


def _is_compatible_paulis(s1: str, s2: str) -> bool:
    for i, j in zip(s1, s2):
        if i == "I" or j == "I":
            continue
        if i != j:
            return False
    return True



from __future__ import annotations

import argparse
import logging
from multiprocessing import cpu_count
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import sinter
from typing_extensions import override

from tqec._cli.subcommands.base import TQECSubCommand
from tqec.compile.convention import ALL_CONVENTIONS
from tqec.gallery.cnot import cnot
from tqec.simulation.plotting.inset import plot_observable_as_inset
from tqec.simulation.simulation import start_simulation_using_sinter
from tqec.utils.enums import Basis
from tqec.utils.noise_model import NoiseModel


class RunExampleTQECSubCommand(TQECSubCommand):
    @staticmethod
    @override
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser: argparse.ArgumentParser = main_parser.add_parser(
            "run-example",
            description=(
                "Runs the full pipeline on the CNOT example. "
                "block graph -> stim circuits and observables -> simulation -> plots"
            ),
        )
        parser.add_argument(
            "--out-dir",
            help="Directory to save the generated outputs.",
            type=Path,
            required=True,
        )
        parser.add_argument(
            "-k",
            help="The scale factors applied to the circuits.",
            nargs="+",
            type=int,
            default=[1, 2, 3],
        )
        parser.add_argument(
            "-p",
            help="The noise levels applied to the simulation.",
            nargs="+",
            type=float,
            default=list(np.logspace(-4, -1, 10)),
        )
        parser.add_argument(
            "--obs-include",
            help=(
                "The observable indices to be included in the circuits. "
                "If not provided, all potential observables will be included."
            ),
            nargs="*",
            type=int,
        )
        parser.add_argument(
            "--convention",
            help="Convention to use.",
            choices=ALL_CONVENTIONS.keys(),
            default="fixed_bulk",
        )
        parser.add_argument(
            "--basis",
            help="Which basis to use for simulation.",
            choices=["X", "Z"],
            default="X",
        )
        parser.set_defaults(func=RunExampleTQECSubCommand.execute)

    @staticmethod
    @override
    def execute(args: argparse.Namespace) -> None:
        # Parse args
        logging.info("Parsing arguments.")
        out_dir: Path = args.out_dir.resolve()
        obs_indices: list[int] = args.obs_include
        convention_name: str = args.convention
        obs_basis = Basis(args.basis.upper())
        ks: list[int] = args.k
        ps: list[float] = args.p

        # Set up the output directories
        logging.info("Setting up output directories.")
        if not out_dir.exists():
            out_dir.mkdir(parents=True)

        simulations_out_dir = out_dir / "simulations"
        simulations_out_dir.mkdir(exist_ok=True)
        plots_out_dir = out_dir / "plots"
        plots_out_dir.mkdir(exist_ok=True)

        logging.info("Generating CNOT block graph and zx graph.")
        block_graph = cnot(obs_basis)
        zx_graph = block_graph.to_zx_graph()

        # observables to a subdirectory
        logging.info("Find the observables.")
        correlation_surfaces = block_graph.find_correlation_surfaces()
        if not obs_indices:
            obs_indices = list(range(len(correlation_surfaces)))
        if max(obs_indices) >= len(correlation_surfaces):
            raise ValueError(
                f"Found {len(correlation_surfaces)} observables,"
                + f"but requested indices up to {max(obs_indices)}."
            )

        convention = ALL_CONVENTIONS[convention_name]

        logging.info("Start the simulation, this might take some time.")
        stats = start_simulation_using_sinter(
            block_graph,
            ks,
            ps,
            NoiseModel.uniform_depolarizing,
            manhattan_radius=2,
            convention=convention,
            observables=[correlation_surfaces[i] for i in obs_indices],
            num_workers=cpu_count(),
            max_shots=1_000_000,
            max_errors=5_000,
            decoders=["pymatching"],
            print_progress=True,
        )
        logging.info("Simulation finished.")

        logging.info("Write plots to %s.", plots_out_dir)
        for i, stat in enumerate(stats):
            with open(
                simulations_out_dir
                / f"{convention}_logical_cnot_result_{obs_basis.value}_observable_{i}.csv",
                "w+",
                encoding="utf-8",
            ) as stats_file:
                stats_file.write(sinter.CSV_HEADER)
                for sub_stat in stat:
                    stats_file.write(sub_stat.to_csv_line())

            fig, ax = plt.subplots()
            sinter.plot_error_rate(
                ax=ax,
                stats=stat,
                x_func=lambda stat: stat.json_metadata["p"],
                group_func=lambda stat: stat.json_metadata["d"],
            )
            plot_observable_as_inset(ax, zx_graph, correlation_surfaces[i])
            ax.grid(axis="both")
            ax.legend()
            ax.loglog()
            ax.set_title(f"{convention} Logical CNOT Error Rate")
            fig.savefig(
                plots_out_dir
                / f"{convention}_logical_cnot_result_{obs_basis.value}_observable_{i}.png"
            )



"""Defines support operations needed to rotate 3D structures upon import to TQEC.

Enabling import of rotated 3D structures is convenient for small computations and
seemingly necessary for large computations. For a small computation, researchers can,
with some pain, rotate the structure back to the original axis or exchange any rotated
cubes if only part of the structure was rotated. But this becomes harder
as the size of the computation grows.

The following descriptions might aid clarity.

In COLLADA files, rotations DO NOT show as "degree" or "radian" rotations applied
to an fixed object. Instead, rotations are encoded directly into the a 4x4 transformation
matrix that looks as follows:
- POS: Position / Translation
- RT: Vector starting at POSITION and shooting RIGHT
- BK: Vector starting at POSITION and shooting BACKWARDS
- UP: Vector starting at POSITION and shooting UP
- US (let's ignore this - won't be using it here): Uniform scaling vector

[RT.x] [UP.x] [BK.x] [POS.x]
[RT.y] [UP.y] [BK.y] [POS.y]
[RT.z] [UP.z] [BK.z] [POS.Z]
[    ] [    ] [    ] [US   ]

The 3x3 submatrix containing RT, UP, BK information will be an identity matrix if object is
unrotated.
Any rotation an user inputs in a software like SketchUp is applied by rotating this matrix
algebraically.
As a result, it is possible to know how much a cube/pipe can be rotated by comparing its
transformation matrix against the original identity matrix (see notes in functions: !).

Additionally, since the names of blocks/pipes in TQEC are tied to the face of the unrotated blocks,
name equivalences can be calculated algebraically using the transformation matrix directly.

"""

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation

from tqec.computation.block_graph import BlockKind, block_kind_from_str
from tqec.utils.exceptions import TQECError
from tqec.utils.position import Direction3D, FloatPosition3D, Position3D
from tqec.utils.scale import round_or_fail


def calc_rotation_angles(
    rotation_matrix: npt.NDArray[np.float32],
) -> npt.NDArray[np.float32]:
    """Calculate the rotation angles of the three rows of matrix (M) from the original X/Y/Z axes.

    Args:
        rotation_matrix: rotation matrix for node, extracted from `.dae` file.

    Returns:
        rotations: the rotation angle for each of the three vectors in M (see notes: !)

    """
    # Placeholder for results
    rotations = np.array([])

    # Define matrix for an unrotated object
    identity = np.identity(3, dtype=int)

    # Calculate rotations
    # ! I think that, technically, this should be done per column (aka column-major)
    # ! but this function is only to confirm rotation validity rather than to transform objects
    # ! per row (aka. row-major) is fine for this
    for i, row in enumerate(rotation_matrix):
        cos_theta = np.dot(identity[i], row) / (np.linalg.norm(identity[i]) * np.linalg.norm(row))
        angle_rad = np.arccos(np.clip(cos_theta, -1.0, 1.0))
        angle_deg = np.degrees(angle_rad)
        rotations = np.append(rotations, [round(angle_deg)])

    return rotations


def get_axes_directions(rotation_matrix: npt.NDArray[np.float32]) -> dict[str, int]:
    """Get up/down multipliers for each row of a rotation matrix.

    Args:
        rotation_matrix: rotation matrix for node.

    Returns:
        axes_directions: up/down multipliers for each axis

    """
    # Placeholder for results
    axes_directions = {"X": 1, "Y": 1, "Z": 1}

    # Loop builds dict with plus/minus direction for each axis
    for i, row in enumerate(rotation_matrix):
        axes_directions["XYZ"[i]] = -1 if np.sum(row) < 0 else 1

    return axes_directions


def rotate_block_kind_by_matrix(
    block_kind: BlockKind, rotation_matrix: npt.NDArray[np.float32]
) -> BlockKind:
    """Multiplies ``rotation_matrix`` with a symbolic vector made from ``block_kind``.

    - ``rotation_matrix`` is NOT rotated: ``block_kind`` untouched
    - ``rotation_matrix`` is rotated: ``block_kind`` rotated accordingly

    Args:
        rotation_matrix: rotation matrix for node.
        block_kind: original kind.

    Returns:
        rotated_kind: rotated kind for the node.

    """
    if str(block_kind) == "PORT":
        return block_kind

    # Placeholder for results
    rotated_name = ""

    # State cultivation blocks: special case - added chars needed to clear loop
    original_name = str(block_kind)[:3] if len(str(block_kind)) > 1 else str(block_kind) + "-!"

    # Loop:
    # - applies transformation encoded in rotate_matrix to vectorised kind
    for row in rotation_matrix:
        entry = ""
        for j, element in enumerate(row):
            entry += abs(int(element)) * original_name[j]
        rotated_name += entry

    # Fails & re-writes for special blocks
    axes_directions = get_axes_directions(rotation_matrix)

    # Reject state cultivation blocks if rotated_name not ends in "!" or axes_directions["Z"]
    # is negative
    if "!" in rotated_name and (not rotated_name.endswith("!") or axes_directions["Z"] < 0):
        raise TQECError(
            f"There is an invalid rotation for {rotated_name.replace('!', '').replace('-', '')} "
            "block.\nCultivation and Y blocks should only allow rotation around Z axis.",
        )
    # Clean kind names for special names
    # State cultivation
    elif "!" in rotated_name:
        rotated_name = str(block_kind)
    # Hadamard
    elif "H" in str(block_kind):
        rotated_name += str(block_kind)[-1]

    # Re-write kind
    rotated_kind = block_kind_from_str(rotated_name)

    return rotated_kind


def get_rotation_matrix(
    rotation_axis: Direction3D,
    counterclockwise: bool = True,
    angle: float = np.pi / 2,
) -> npt.NDArray[np.float32]:
    """Get the rotation matrix for a given axis and rotation.

    Args:
        rotation_axis: axis to rotate around.
        counterclockwise: whether to rotate counterclockwise.
        angle: rotation angle in radians.

    Returns:
        The rotation matrix.

    """
    rot_vec = np.array([0, 0, 0])
    rot_vec[rotation_axis.value] = 1 if counterclockwise else -1
    return np.array(Rotation.from_rotvec(rot_vec * angle).as_matrix(), dtype=np.float32)


def rotate_position_by_matrix(
    position: Position3D,
    rotation_matrix: npt.NDArray[np.float32],
) -> Position3D:
    """Rotates a cube by a given rotation matrix and returns the new position.

    Note that we rotate the position of the cube based on its center, while the
    position is based on its corner. Therefore, a cube at (0, 0, 0) rotated by
    90 degrees around the x-axis will be at (0, -1, 0).

    Args:
        position: cube position to rotate.
        rotation_matrix: rotation matrix.

    Returns:
        The rotated position.

    Raises:
        TQECError: if the rotated position is not integer.

    """
    rotation = Rotation.from_matrix(rotation_matrix)
    center_pos = [i + 0.5 for i in position.as_tuple()]
    rotated_center = rotation.apply(center_pos)
    rotated_corner = [round_or_fail(float(i) - 0.5) for i in rotated_center]
    return Position3D(*rotated_corner)


def rotate_on_import(
    rotation_matrix: npt.NDArray[np.float32],
    translation_matrix: npt.NDArray[np.float32],
    scale_matrix: npt.NDArray[np.float32],
    kind: BlockKind,
) -> tuple[FloatPosition3D, BlockKind]:
    """Update the kind of an incoming block when rotated.

    The block kind is only updated when its translation matrix indicates the original block has been
    rotated, rejecting any invalid rotation in the process.

    Args:
        rotation_matrix: rotation matrix of the incoming block.
        translation_matrix: translation matrix of the incoming block.
        scale_matrix: scaling factor of the incoming block.
        kind: kind of the original block that was rotated / requires rotation.

    Raises:
        TQECError: if an invalid rotation is provided.

    Returns:
        A tuple containing two entries:

        - translation: An updated translation matrix
        - kind: An updated kind that factors the rotation into the kind itself

    """
    # Calculate rotation
    rotation_angles = calc_rotation_angles(rotation_matrix)

    # Reject invalid rotations for all other cubes/pipes:
    if (
        # Any rotation with angle not an integer multiply of 90 degrees: partially rotated
        # block/pipe
        any([int(angle) not in [0, 90, 180] for angle in rotation_angles])
        # At least 1 * 180-deg or 2 * 90-deg rotation to avoid dimensional collapse
        # (A single 90-deg rotation would put the rotated vector on the plane made by the other
        # two axes)
        or sum([angle for angle in rotation_angles]) < 180
    ):
        raise TQECError(  # pragma: no cover
            f"There is an invalid rotation for {kind} block at "
            f"position {FloatPosition3D(*translation_matrix.tolist())}."
        )

    # Rotate node name
    # Calculate rotated kind and directions for all axes in case it is needed
    kind = rotate_block_kind_by_matrix(kind, rotation_matrix)

    # Shift nodes slightly according to rotation
    translation = FloatPosition3D(*translation_matrix + rotation_matrix.dot(scale_matrix))

    # Return revised data
    return translation, kind


def adjust_hadamards_direction(kind: BlockKind) -> BlockKind:
    """Inverts the direction of any Hadamard pipe.

    This function inverts the direction of any "h" pipe when called as applicable (when the pipe
    runs in the negative direction on any given axis) by exchanging the kind for the corresponding
    pair on the given axis.

    Args:
        kind: the original "h" kind.

    Returns:
        the updated (inverse) "h" kind.

    """
    # List of hadamard equivalences
    hdm_equivalences = {"ZXOH": "XZOH", "XOZH": "ZOXH", "OXZH": "OZXH"}

    # Match to equivalent block given direction
    if str(kind) in hdm_equivalences.keys():
        kind = block_kind_from_str(hdm_equivalences[str(kind)])
    else:
        inv_equivalences = {value: key for key, value in hdm_equivalences.items()}
        kind = block_kind_from_str(inv_equivalences[str(kind)])

    # Return revised kind
    return kind



import numpy
import numpy.typing as npt

from tqec.utils.exceptions import TQECError


def to2dlist(array: npt.NDArray[numpy.int_]) -> list[list[int]]:
    """Transform a 2D numpy array into a 2D array using Python list."""
    if len(array.shape) != 2:
        raise TQECError(f"Cannot transform a {len(array.shape)}-dimensional array to 2D.")
    return [[int(v) for v in line] for line in array]



"""Provides helper functions to add observables to circuits."""

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Protocol

import stim

from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.circuit.qubit import GridQubit
from tqec.compile.observables.abstract_observable import (
    AbstractObservable,
    CubeWithArms,
    PipeWithArms,
    PipeWithObservableBasis,
)
from tqec.templates.layout import LayoutTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import PlaquetteShape2D, Position3D
from tqec.utils.scale import round_or_fail


@dataclass(frozen=True)
class Observable:
    """Logical observable consisting of a list of measurements.

    Attributes:
        observable_index: The index of the observable in the circuit.
        measured_qubits: The qubits whose measurements will be included in the
            observable.
        measurement_offsets: The offsets of the measurements in the circuit.
            This should have the same length as `measured_qubits` and
            contain strictly negative values.

    """

    observable_index: int
    measured_qubits: list[GridQubit]
    measurement_offsets: list[int]

    def __post_init__(self) -> None:
        if len(self.measured_qubits) != len(self.measurement_offsets):
            raise TQECError(
                "The number of measured qubits and measurement offsets must be the same."
            )
        if any(m >= 0 for m in self.measurement_offsets):
            raise TQECError("Expected strictly negative measurement offsets.")

    def to_instruction(self) -> stim.CircuitInstruction:
        """Return the ``stim`` instruction reprensented by ``self``."""
        return stim.CircuitInstruction(
            "OBSERVABLE_INCLUDE",
            [stim.target_rec(offset) for offset in sorted(self.measurement_offsets)],
            [self.observable_index],
        )


class ObservableComponent(Enum):
    BOTTOM_STABILIZERS = "bottom_stabilizers"
    TOP_READOUTS = "top_readouts"
    REALIGNMENT = "realignment"


Coordinates2D = tuple[float, float] | tuple[int, int]


class CubeObservableQubitsBuilder(Protocol):
    """Build the qubits whose measurements will be included in the logical observable that is
    supported by the cube.

    This calculates the coordinates of these data qubits in the local coordinate system.

    """

    def __call__(self, shape: PlaquetteShape2D, cube: CubeWithArms) -> Sequence[Coordinates2D]:
        """Build the qubit coordinates whose measurement will be included in the observable."""
        ...


class PipeObservableQubitsBuilder(Protocol):
    """Build the qubits whose measurements will be included in the logical observable that is
    supported by the pipe.

    This calculates the coordinates of that data qubit in the local coordinate system of the cube at
    the head of the pipe.

    """

    def __call__(self, shape: PlaquetteShape2D, pipe: PipeWithArms) -> Sequence[Coordinates2D]:
        """Build the qubit coordinates whose measurement will be included in the observable."""
        ...


class TemporalPipeObservableQubitsBuilder(Protocol):
    """Build the qubits whose measurements will be included in the logical observable that is
    supported by the temporal pipe.

    This calculates the coordinates of that data qubit in the local coordinate system of the cube at
    the head of the pipe.

    """

    def __call__(
        self, shape: PlaquetteShape2D, pipe: PipeWithObservableBasis
    ) -> Sequence[Coordinates2D]:
        """Build the qubit coordinates whose measurement will be included in the observable."""
        ...


@dataclass
class ObservableBuilder:
    """Compute the qubits whose measurements will be included in the logical observable.

    The builders can include the qubits that are not in the circuit like qubits in the
    scretched stabilizers to simplify the calculation. The qubits that are not
    measured in the circuit will be ignored when calling
    ``get_observable_with_measurement_records``.

    """

    cube_top_readouts_builder: CubeObservableQubitsBuilder
    pipe_top_readouts_builder: PipeObservableQubitsBuilder
    cube_bottom_stabilizers_builder: CubeObservableQubitsBuilder
    pipe_bottom_stabilizers_builder: PipeObservableQubitsBuilder
    pipe_temporal_hadamard_builder: TemporalPipeObservableQubitsBuilder

    def build(
        self,
        k: int,
        template: LayoutTemplate,
        observable: AbstractObservable,
        component: ObservableComponent,
    ) -> set[GridQubit]:
        """Compute the qubits whose measurements will be included in the observable.

        Args:
            k: The scaling parameter.
            template: The layout template of the circuit.
            observable: The abstract observable that builds the observable qubits from.
            component: The component of the observable to compute.

        Returns:
            A set of qubits whose measurements will be included in the logical observable.
            The qubits are in the global coordinate system of the circuit.

        """
        shape = template.element_shape(k)
        obs_qubits: set[GridQubit] = set()

        # The stabilizer measurements that will be added to the end of the
        # first layer of circuits at z.
        if component == ObservableComponent.BOTTOM_STABILIZERS:
            for cube in observable.bottom_stabilizer_cubes:
                obs_qubits.update(
                    self.transform_coords_into_grid(
                        k,
                        template,
                        self.cube_bottom_stabilizers_builder(shape, cube),
                        cube.cube.position,
                    )
                )
            for pipe in observable.bottom_stabilizer_pipes:
                obs_qubits.update(
                    self.transform_coords_into_grid(
                        k,
                        template,
                        self.pipe_bottom_stabilizers_builder(shape, pipe),
                        pipe.pipe.u.position,
                    )
                )
        # The readouts that will be added to the end of the last layer of circuits at z.
        elif component == ObservableComponent.TOP_READOUTS:
            for cube in observable.top_readout_cubes:
                obs_qubits.update(
                    self.transform_coords_into_grid(
                        k,
                        template,
                        self.cube_top_readouts_builder(shape, cube),
                        cube.cube.position,
                    )
                )
            for pipe in observable.top_readout_pipes:
                obs_qubits.update(
                    self.transform_coords_into_grid(
                        k,
                        template,
                        self.pipe_top_readouts_builder(shape, pipe),
                        pipe.pipe.u.position,
                    )
                )
        else:  # component == ObservableComponent.REALIGNMENT
            for hadamard_pipe in observable.temporal_hadamard_pipes:
                obs_qubits.update(
                    self.transform_coords_into_grid(
                        k,
                        template,
                        self.pipe_temporal_hadamard_builder(shape, hadamard_pipe),
                        hadamard_pipe.pipe.u.position,
                    )
                )
        return obs_qubits

    @staticmethod
    def transform_coords_into_grid(
        k: int,
        template: LayoutTemplate,
        local_coords: Iterable[tuple[float, float] | tuple[int, int]],
        block_position: Position3D,
    ) -> set[GridQubit]:
        """Transform local coordinates at a block to the global coordinates in the circuit.

        When calculating the coordinates of the measurement qubits, we use a local
        coordinate system in the individual blocks. The top-left corner of the
        block starts at (0, 0) and each column is separated by 0.5. That is, all
        the data qubits are placed at the integer coordinates while all the
        measurement qubits are placed at the half-integer coordinates.

        This convention helps reducing the number of arguments needed to pass around
        and simplifies the calculation of the qubit coordinates. The global coordinates
        are calculated by offsetting the local coordinates by the global position of
        the block and accounting for the ``Template.default_increments``.

        """
        block_shape = template.element_shape(k)
        template_increments = template.get_increments()
        width = block_shape.x * template_increments.x
        height = block_shape.y * template_increments.y
        transformed_coords: set[GridQubit] = set()
        for coords in local_coords:
            x = block_position.x * width + round_or_fail((coords[0] - 0.5) * template_increments.x)
            y = block_position.y * height + round_or_fail((coords[1] - 0.5) * template_increments.y)
            transformed_coords.add(GridQubit(x, y))
        return transformed_coords


def get_observable_with_measurement_records(
    qubits: set[GridQubit],
    measurement_records: MeasurementRecordsMap,
    observable_index: int,
    ignore_qubits_with_no_measurement: bool = True,
) -> Observable:
    """Calculate the measurement offsets of the observable qubits measured at the end of the circuit
    and construct the observable.

    Args:
        qubits: The qubits whose measurements will be included in the observable.
        measurement_records: The measurement records of the qubits in a circuit.
        observable_index: The index of the observable.
        ignore_qubits_with_no_measurement: Whether to skip qubits that are not
            measured in the circuit. If set to False, an exception will be raised
            if any of the qubits is not measured.

    Returns:
        The logical observable.

    """
    if not ignore_qubits_with_no_measurement and any(
        len(measurement_records.mapping.get(q, [])) == 0 for q in qubits
    ):
        raise TQECError(
            "Some qubits are not measured in the circuit. Set "
            "ignore_qubits_with_no_measurement to True to ignore them."
        )

    measured_qubits = [
        q
        for q in qubits
        # Ignore those qubits that are not measured in the circuit.
        # This is required because the some observable builders
        # include the qubits that are not in the circuit like qubits
        # in the scretched stabilizers to simplify the calculation.
        if q in measurement_records
    ]
    measurement_offsets = [measurement_records[q][-1] for q in measured_qubits]
    return Observable(observable_index, measured_qubits, measurement_offsets)
